"""
Utils for Otter Export exporterss
"""

import re
import copy
import nbformat

from ...utils import get_source

BEGIN_QUESTION_REGEX = r"<!--\s*BEGIN QUESTION\s*-->"
END_QUESTION_REGEX = r"<!--\s*END QUESTION\s*-->"
NEW_PAGE_REGEX = r"<!--\s*NEW PAGE\s*-->"
NEW_PAGE_MARKER = "#newpage"
NEW_PAGE_CELL_SOURCE = f"<!-- {NEW_PAGE_MARKER} -->"
NEW_PAGE_CLASS_NAME = "otter-page-break-after"

def has_begin(line):
    """
    Returns whether a string contains a begin question comment

    A begin question comment is an HTML comment on a single line that denotes the beginning of an 
    export block. The begin question comment looks like:

    .. code-block:: html

        <!-- BEGIN QUESTION -->
    
    Args:
        line (``str``): the line to search
    
    Returns:
        ``bool``: whether the line contains a substring matching the begin question regex
    """
    return bool(re.search(BEGIN_QUESTION_REGEX, line, flags=re.IGNORECASE))

def has_end(line):
    """
    Returns whether a string contains an end question comment

    An end question comment is an HTML comment on a single line that denotes the end of an export 
    block. The begin question comment looks like:

    .. code-block:: html

        <!-- END QUESTION -->
    
    Args:
        line (``str``): the line to search
    
    Returns:
        ``bool``: whether the line contains a substring matching the end question regex
    """
    return bool(re.search(END_QUESTION_REGEX, line, flags=re.IGNORECASE))

def sub_end_for_new_page(line):
    """
    Subsitutes an end question comment for a newpage comment

    The end question HTML comment (cf. ``otter.export.filter.has_end``) is replaced with the following
    HTML comment to indicate a pagebreak in the LaTeX template.

    .. code-block:: html

        <!-- #newpage -->

    Args:
        ``line``: the line to substitute in

    Returns:
        ``str``: the line with the end question match substituted for the newpage comment
    """
    return re.sub(END_QUESTION_REGEX, NEW_PAGE_CELL_SOURCE, line)

def notebook_pdf_generator(nb):
    """
    A generator that takes in a notebook ``nb`` with HTML comments for filtering and splits this
    notebook up into each filtered block, yielding a complete notebook for each chunk. Used for 
    implementing pagebreaks in PDFs via HTML.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to be exported

    Yields:
        ``nbformat.NotebookNode``: a complete notebook containing a single filtered block
    """
    dummy_nb = copy.copy(nb)
    dummy_nb.cells = []

    all_cells, subnb_cells = [], []
    for cell in nb.cells:
        source = get_source(cell)

        if NEW_PAGE_CELL_SOURCE in "\n".join(source):
            for i, line in enumerate(source):
                if NEW_PAGE_CELL_SOURCE in line:
                    break

            c1, c2 = nbformat.v4.new_markdown_cell(), nbformat.v4.new_markdown_cell()
            c1.source, c2.source = "\n".join(source[:i+1]), "\n".join(source[i+1:])

            subnb_cells.append(c1)
            all_cells.append(subnb_cells)
            subnb_cells = []
            subnb_cells.append(c2)
            
        else:
            subnb_cells.append(cell)
    
    all_cells.append(subnb_cells)
    
    for subnb_cells in all_cells:
        dummy_nb.cells = subnb_cells
        yield dummy_nb



#################################
##### Tests for otter grade #####
#################################

import pandas as pd
import os
import re
import shutil
import subprocess
import zipfile

from glob import glob
from subprocess import PIPE

from otter.generate import main as generate
from otter.generate import utils
from otter.grade import main as grade

from . import TestCase


TEST_FILES_PATH = "test/test-grade/"


class TestGrade(TestCase):
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        
        create_image_cmd = ["make", "docker-grade-test"]
        subprocess.run(create_image_cmd, check=True)

        shutil.copy("otter/grade/Dockerfile", "otter/grade/old-Dockerfile")
        with open("otter/grade/Dockerfile", "r+") as f:
            lines = f.readlines()

            idx = max([i if "ARG" in lines[i] else -1 for i in range(len(lines))])
            lines.insert(idx + 1, "ADD otter-grader /home/otter-grader\n")

            f.seek(0)
            f.write("".join(lines))

        cls.generate_autograder_zip(pdfs=True)
        
    def setUp(self):
        """
        Load in point values
        """
        self.test_points = {}
        for test_file in glob(TEST_FILES_PATH + "tests/*.py"):
            env = {}
            with open(test_file) as f:
                exec(f.read(), env)
            self.test_points[env['test']['name']] = env['test']['points']
        return super().setUp()

    @staticmethod
    def generate_autograder_zip(pdfs=False):
        generate(
            tests_dir = TEST_FILES_PATH + "tests", 
            requirements = TEST_FILES_PATH + "requirements.txt", 
            output_path = TEST_FILES_PATH + "autograder.zip",
            config = TEST_FILES_PATH + "otter_config.json" if pdfs else None,
            no_environment = True,
        )
        with zipfile.ZipFile(TEST_FILES_PATH + "autograder.zip", "a") as zip_ref:
            utils.zip_folder(zip_ref, os.getcwd(), exclude=[".git", "logo", "test", "dist", "build", "otter_grader.egg-info"])

    def test_docker(self):
        """
        Check that we have the right container installed and that docker is running
        """
        # use docker image inspect to see that the image is installed and tagged as otter-grader
        inspect = subprocess.run(["docker", "image", "inspect", "otter-test"], stdout=PIPE, stderr=PIPE)

        # assert that it didn't fail, it will fail if it is not installed
        self.assertEqual(len(inspect.stderr), 0, inspect.stderr.decode("utf-8"))

    def test_timeout(self):
        """
        Check that the notebook `20s.ipynb` is killed due to exceeding the defined timeout.
        """
        with self.assertRaises(Exception) as e:
            grade(
                path=TEST_FILES_PATH + "timeout/",
                output_dir="test/",
                autograder=TEST_FILES_PATH + "autograder.zip",
                containers=5,
                image="otter-test",
                timeout=40,
                debug=True,
                verbose=True,
            )
        self.assertEqual(str(e.exception),"Executing 'test/test-grade/timeout/20s.ipynb' in docker container failed! Exit code: 137")

    def test_network(self):
        """
        Check that the notebook `network.ipynb` is unable to do some network requests with disabled networking
        """

        grade(
            path = TEST_FILES_PATH + "network/",
            output_dir = "test/",
            autograder = TEST_FILES_PATH + "autograder.zip",
            containers = 5,
            image = "otter-test",
            pdfs = True,
            no_network=True
        )
        df_test = pd.read_csv("test/final_grades.csv")

        # sort by filename
        df_test = df_test.sort_values("file").reset_index(drop=True)

        for _, row in df_test.iterrows():
            for test in self.test_points:
                if row['file'] == 'network.ipynb' and ('q2' in test or 'q3' in test):
                    self.assertEqual(row[test],0,"{} supposed to fail {} but passed".format(row["file"], test))
                else:
                    self.assertEqual(row[test], self.test_points[test], "{} supposed to pass {} but failed".format(row["file"], test))

    def test_notebooks_with_pdfs(self):
        """
        Check that the example of 100 notebooks runs correctely locally.
        """

        # grade the 100 notebooks
        grade(
            path = TEST_FILES_PATH + "notebooks/", 
            output_dir = "test/",
            autograder = TEST_FILES_PATH + "autograder.zip",
            containers = 5,
            image = "otter-test",
            pdfs = True,
            debug = True,
            verbose = True
        )

        # read the output and expected output
        df_test = pd.read_csv("test/final_grades.csv")

        # sort by filename
        df_test = df_test.sort_values("file").reset_index(drop=True)
        df_test["failures"] = df_test["file"].apply(lambda x: [int(n) for n in re.split(r"\D+", x) if len(n) > 0])

        # add score sum cols for tests
        for test in self.test_points:
            test_cols = [l for l in df_test.columns if bool(re.search(fr"\b{test}\b", l))]
            df_test[test] = df_test[test_cols].sum(axis=1)

        # check point values
        for _, row in df_test.iterrows():
            for test in self.test_points:
                if int(re.sub(r"\D", "", test)) in row["failures"]:
                    # q6.py has all_or_nothing set to False, so if the hidden tests fail you should get 2.5 points
                    if "6H" in row["file"] and "q6" == test:
                        self.assertEqual(row[test], 2.5, "{} supposed to fail {} but passed".format(row["file"], test))
                    else:
                        self.assertEqual(row[test], 0, "{} supposed to fail {} but passed".format(row["file"], test))
                else:
                    self.assertEqual(row[test], self.test_points[test], "{} supposed to pass {} but failed".format(row["file"], test))

        self.assertTrue(os.path.exists("test/submission_pdfs"), "PDF folder is missing")

        # check that an pdf exists for each submission
        dir1_contents, dir2_contents = (
            [os.path.splitext(f)[0] for f in os.listdir(TEST_FILES_PATH + "notebooks/") if not (os.path.isdir(os.path.join(TEST_FILES_PATH + "notebooks/", f)))],
            [os.path.splitext(f)[0] for f in os.listdir("test/submission_pdfs") if not (os.path.isdir(os.path.join("test/submission_pdfs", f)))],
        )
        self.assertEqual(sorted(dir1_contents), sorted(dir2_contents), f"'{TEST_FILES_PATH}notebooks/' and 'test/submission_pdfs' have different contents")

    def tearDown(self) -> None:
        # remove the extra output
        cleanup_command = ["rm", "-rf", "test/final_grades.csv", "test/submission_pdfs", "test/final_grades.csv"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(cleanup.stderr), 0, cleanup.stderr.decode("utf-8"))

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        create_image_cmd = ["make", "cleanup-docker-grade-test"]
        subprocess.run(create_image_cmd, check=True)
        if os.path.exists("otter/grade/old-Dockerfile"):
            os.remove("otter/grade/Dockerfile")
            shutil.move("otter/grade/old-Dockerfile", "otter/grade/Dockerfile")
        if os.path.exists(TEST_FILES_PATH + "autograder.zip"):
            os.remove(TEST_FILES_PATH + "autograder.zip")
        # prune images
        grade(prune=True, force=True)



"""Local grading of submissions in Docker containers for Otter-Grader"""

import os
import pandas as pd
import re

from .containers import launch_grade
from .utils import merge_csv, prune_images

from ..utils import assert_path_exists


_ALLOWED_EXTENSIONS = ["ipynb", "py", "Rmd", "R", "r"]


def main(*, path="./", output_dir="./", autograder="./autograder.zip", containers=None, 
         ext="ipynb", no_kill=False, debug=False, zips=False, image="ucbdsinfra/otter-grader", 
         pdfs=False, verbose=False, prune=False, force=False, timeout=None, no_network=False):
    """
    Runs Otter Grade

    Grades a directory of submissions in parallel Docker containers. Results are outputted as a CSV file
    called ``final_grades.csv``. If ``prune`` is ``True``, Otter's dangling grading images are pruned 
    and the program exits.

    Args:
        path (``str``): path to directory of submissions
        output_dir (``str``): directory in which to write ``final_grades.csv``
        autograder (``str``): path to Otter autograder configuration zip file
        containers (``int``): number of containers to run in parallel
        ext (``str``): the submission file extension for globbing
        no_kill (``bool``): whether to keep containers after grading is finished
        debug (``bool``): whether to print the stdout of each container
        zips (``bool``): whether the submissions are Otter-exported zip files
        image (``str``): base image from which to build grading image
        pdfs (``bool``): whether to copy notebook PDFs out of the containers
        verbose (``bool``): whether to log status messages to stdout
        prune (``bool``): whether to prune the grading images; if true, no grading is performed
        force (``bool``): whether to force-prune the images (do not ask for confirmation)
        timeout (``int``): timeout in seconds for each container
        no_network (``bool``): whether to disable networking in the containers

    Raises:
        ``AssertionError``: if invalid arguments are provided
    """
    if prune:
        prune_images(force=force)
        return

    # check file paths
    assert_path_exists([
        (path, True),
        (output_dir, True),
        (autograder, False),
    ])

    if ext not in _ALLOWED_EXTENSIONS:
        raise ValueError(f"Invalid submission extension specified: {ext}")

    if verbose:
        print("Launching docker containers...")

    #Docker
    grade_dfs = launch_grade(autograder,
        submissions_dir=path,
        verbose=verbose,
        num_containers=containers,
        ext=ext,
        no_kill=no_kill,
        output_path=output_dir,
        debug=debug,
        zips=zips,
        image=image,
        pdfs=pdfs,
        timeout=timeout,
        network=not no_network,
    )

    if verbose:
        print("Combining grades and saving...")

    # Merge Dataframes
    output_df = merge_csv(grade_dfs)
    cols = output_df.columns.tolist()
    output_df = output_df[cols[-1:] + cols[:-1]]

    # write to CSV file
    output_df.to_csv(os.path.join(output_dir, "final_grades.csv"), index=False)



"""CLI for converting v0-formatted notebooks to v1 format"""

import click
import nbformat

from .notebook_transformer import get_transformed_cells
from ..constants import NB_VERSION
from ...utils import remove_cell_ids


@click.command()
@click.argument("src", type=click.Path(exists=True, dir_okay=False))
@click.argument("dst", type=click.Path())
def main(src, dst):
    """
    Convert the Assign master notebook SRC from v0 format to v1 format and write the resulting
    notebook at DST.
    """
    nb = nbformat.read(src, as_version=NB_VERSION)
    cells = get_transformed_cells(nb["cells"])
    nb["cells"] = cells
    remove_cell_ids(nb)
    nbformat.write(nb, dst)


if __name__ == "__main__":
    main()



"""Execution of a submission through log deserialization"""

import re

from IPython.core.inputsplitter import IPythonInputSplitter
from unittest import mock

from ..utils import get_variable_type


def execute_log(nb, log, check_results_list_name="check_results_secret", initial_env=None, 
                ignore_errors=False, cwd=None, test_dir=None, variables=None):
    """
    Execute a notebook from logged environments and return the global environment that results.

    If ``ignore_errors`` is true, exceptions are swallowed.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to execute
        log (``otter.check.logs.Log``): log from notebook execution
        check_results_list_name (``str``, optional): the name of the list to collect check results in
        initial_env (``str``, optional): name of initial environment
        ignore_errors (``bool``, optional): whether exceptions should be ignored
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        variables (``dict``, optional): map of variable names -> type string to check type of deserialized
            object to prevent arbitrary code from being put into the environment
    
    Results:
        ``dict``: global environment resulting from executing all code of the input notebook
    """
    if initial_env:
        global_env = initial_env.copy()
    else:
        global_env = {}

    if test_dir:
        source = f"import otter\ngrader = otter.Notebook(\"{test_dir}\")\n"
    else:
        source = f"import otter\ngrader = otter.Notebook()\n"
    
    if cwd:
        source +=  f"import sys\nsys.path.append(\"{cwd}\")\n"

    logged_questions = []
    m = mock.mock_open()
    with mock.patch("otter.Notebook._log_event", m):
        exec(source, global_env)

        for cell in nb['cells']:
            if cell['cell_type'] == 'code':
                # transform the input to executable Python
                # FIXME: use appropriate IPython functions here
                isp = IPythonInputSplitter(line_input_checker=False)
                
                code_lines = []
                cell_source_lines = cell['source']
                source_is_str_bool = False
                if isinstance(cell_source_lines, str):
                    source_is_str_bool = True
                    cell_source_lines = cell_source_lines.split('\n')

                # only execute import statements
                cell_source_lines = [re.sub(r"^\s+", "", l) for l in cell_source_lines if "import" in l]                                
                
                for line in cell_source_lines:
                    try:
                        exec(line, global_env)

                    except:
                        if not ignore_errors:
                            raise

        for entry in log.question_iterator():
            shelf = entry.unshelve(global_env)

            if variables is not None:
                for k, v in shelf.items():
                    full_type = get_variable_type(v)
                    if not (k in variables and variables[k] == full_type):
                        del shelf[k]
                        print(f"Found variable of different type than expected: {k}")

            global_env.update(shelf)
            global_env[check_results_list_name].append(
                global_env["grader"].check(entry.question, global_env=global_env))
            logged_questions.append(entry.question)

    print("Questions executed from log: {}".format(", ".join(logged_questions)))
    
    return global_env



"""Autograder configuration generator for Otter-Grader"""

import json
import os
import pathlib
import pkg_resources
import shutil
import tempfile
import yaml
import zipfile

from glob import glob
from jinja2 import Template
from subprocess import PIPE

from .token import APIClient
from .utils import zip_folder

from ..plugins import PluginCollection
from ..run.run_autograder.constants import DEFAULT_OPTIONS
from ..utils import load_default_file


TEMPLATE_DIR = pkg_resources.resource_filename(__name__, "templates")
MINICONDA_INSTALL_URL = "https://repo.anaconda.com/miniconda/Miniconda3-py38_4.10.3-Linux-x86_64.sh"
OTTER_ENV_NAME = "otter-env"
OTTR_BRANCH = "1.1.3"  # this should match a release tag on GitHub


def main(*, tests_dir="./tests", output_path="autograder.zip", config=None, no_config=False, 
         lang="python", requirements=None, no_requirements=False, overwrite_requirements=False, 
         environment=None, no_environment=False, username=None, password=None, token=None, files=[], 
         assignment=None, plugin_collection=None):
    """
    Runs Otter Generate

    Args:
        tests_dir (``str``): path to directory of test files for this assignment
        output_path (``str``): the path at which to write the output zip file
        config (``str``): path to an Otter configuration JSON file
        no_config (``bool``): disables auto-inclusion of Otter config file at ./otter_config.json
        lang (``str``): the language of the assignment; one of ``["python", "r"]``
        requirements (``str``): path to a Python or R requirements file for this assignment
        no_requirements (``bool``): disables auto-inclusion of requirements file at ./requirements.txt
        overwrite_requirements (``bool``): whether to overwrite the default requirements instead of
            adding to them
        environment (``str``): path to a conda environment file for this assignment
        no_environment (``bool``): whether ``./evironment.yml`` should be automatically checked if 
            ``environment`` is unspecified
        username (``str``): a username for Gradescope for generating a token
        password (``str``): a password for Gradescope for generating a token
        token (``str``): a token for Gradescope
        files (``list[str]``): list of file paths to add to the zip file
        assignment (``otter.assign.assignment.Assignment``, optional): the assignment configurations
            if used with Otter Assign
        **kwargs: ignored kwargs (a remnant of how the argument parser is built)

    Raises:
        ``FileNotFoundError``: if the specified Otter configuration JSON file could not be found
        ``ValueError``: if the configurations specify a Gradescope course ID or assignment ID but not
            both
    """
    # read in otter_config.json
    if config is None and os.path.isfile("otter_config.json") and not no_config:
        config = "otter_config.json"

    if config is not None and not os.path.isfile(config):
        raise FileNotFoundError(f"Could not find otter configuration file {config}")

    if config:
        with open(config, encoding="utf-8") as f:
            otter_config = json.load(f)
    else:
        otter_config = {}

    # if an empty/null token is specified, delete it
    if "token" in otter_config and not otter_config["token"]:
        otter_config.pop("token")

    # ensure that a token is present if necessary
    if "token" not in otter_config and token is not None:
        otter_config["token"] = token
    
    elif "token" not in otter_config and "course_id" in otter_config and "assignment_id" in otter_config:
        client = APIClient()
        if username is not None and password is not None:
            client.log_in(username, password)
            token = client.token
        else:
            token = client.get_token()
        otter_config["token"] = token
    
    elif ("course_id" in otter_config) ^ ("assignment_id" in otter_config):
        raise ValueError(f"Otter config contains 'course_id' or 'assignment_id' but not both")

    options = DEFAULT_OPTIONS.copy()
    options.update(otter_config)

    # update language
    options["lang"] = options.get("lang", lang.lower())

    template_dir = os.path.join(TEMPLATE_DIR, options["lang"])

    templates = {}
    for fn in os.listdir(template_dir):
        fp = os.path.join(template_dir, fn)
        if os.path.isfile(fp): # prevents issue w/ finding __pycache__ in template dirs
            with open(fp) as f:
                templates[fn] = Template(f.read())

    template_context = {
        "autograder_dir": options['autograder_dir'],
        "otter_env_name": OTTER_ENV_NAME,
        "miniconda_install_url": MINICONDA_INSTALL_URL,
        "ottr_branch": OTTR_BRANCH,
        "channel_priority_strict": options["channel_priority_strict"],
    }

    if plugin_collection is None:
        plugin_collection = PluginCollection(otter_config.get("plugins", []), None, {})
    else:
        plugin_collection.add_new_plugins(otter_config.get("plugins", []))
    
    plugin_collection.run("during_generate", otter_config, assignment)

    # create tmp directory to zip inside
    with tempfile.TemporaryDirectory() as td:

        # try:
        # copy tests into tmp
        test_dir = os.path.join(td, "tests")
        os.mkdir(test_dir)
        pattern = ("*.py", "*.[Rr]")[options["lang"] == "r"]
        for file in glob(os.path.join(tests_dir, pattern)):
            shutil.copy(file, test_dir)

        # open requirements if it exists
        with load_default_file(
            requirements, 
            f"requirements.{'R' if options['lang'] == 'r' else 'txt'}", 
            default_disabled=no_requirements
        ) as reqs:
            template_context["other_requirements"] = reqs if reqs is not None else ""

        template_context["overwrite_requirements"] = overwrite_requirements

        # open environment if it exists
        # unlike requirements.txt, we will always overwrite, not append by default
        with load_default_file(environment, "environment.yml", default_disabled=no_environment) as env_contents:
            template_context["other_environment"] = env_contents
            if env_contents is not None:
                data = yaml.safe_load(env_contents)
                data['name'] = template_context["otter_env_name"]
                template_context["other_environment"] = yaml.safe_dump(data, default_flow_style=False)
  
        rendered = {}
        for fn, tmpl in templates.items():
            rendered[fn] = tmpl.render(**template_context)
        
        if os.path.exists(output_path):
            os.remove(output_path)

        with zipfile.ZipFile(output_path, mode="w") as zf:
            for fn, contents in rendered.items():
                zf.writestr(fn, contents)

            test_dir = "tests"
            pattern = ("*.py", "*.[Rr]")[options["lang"] == "r"]
            for file in glob(os.path.join(tests_dir, pattern)):
                zf.write(file, arcname=os.path.join(test_dir, os.path.basename(file)))
            
            zf.writestr("otter_config.json", json.dumps(otter_config, indent=2))

            # copy files into tmp
            if len(files) > 0:
                for file in files:
                    full_fp = os.path.abspath(file)
                    assert os.getcwd() in full_fp, f"{file} is not in a subdirectory of the working directory"
                    if os.path.isfile(full_fp):
                        zf.write(file, arcname=os.path.join("files", file))
                    elif os.path.isdir(full_fp):
                        zip_folder(zf, full_fp, prefix="files")
                    else:
                        raise ValueError(f"Could not find file or directory '{full_fp}'")
    
    if assignment is not None:
        assignment._otter_config = otter_config



"""
Classes for working with test files
"""

import os
import math
import json
import pprint
import pickle

from collections import namedtuple

from .abstract_test import TestCase, TestCaseResult
from .metadata_test import NotebookMetadataOKTestFile
from .ok_test import OKTestFile
from .ottr_test import OttrTestFile

from ..check.logs import QuestionNotInLogException
from ..run.run_autograder.constants import DEFAULT_OPTIONS


GradingTestCaseResult = namedtuple(
    "GradingTestCaseResult", 
    ["name", "score", "possible", "hidden", "incorrect", "test_case_result", "test_file"]
)


class GradingResults:
    """
    Stores and wrangles test result objects
    
    Initialize with a list of ``otter.test_files.abstract_test.TestFile`` subclass objects and 
    this class will store the results as named tuples so that they can be accessed/manipulated easily. 
    Also contains methods to put the results into a nice ``dict`` format or into the correct format 
    for Gradescope.

    Args:
        results (``list`` of ``TestFile``): the list of test file objects summarized in this grade
    
    Attributes:
        results (``dict``): maps test names to ``GradingTestCaseResult`` named tuples containing the 
            test result information
        output (``str``): a string to include in the output field for Gradescope
        all_hidden (``bool``): whether all results should be hidden from the student on Gradescope
        tests (``list`` of ``str``): list of test names according to the keys of ``results``
    """
    def __init__(self, test_files):
        self._plugin_data = {}
        self.results = {tf.name: tf for tf in test_files}
        # self.results = {}
        self.output = None
        self.all_hidden = False

    def __repr__(self):
        return self.summary()

    @classmethod
    def from_ottr_json(cls, ottr_output):
        """
        Creates a ``GradingResults`` object from the JSON output of Ottr.

        Args:
            ottr_output (``str``): the JSON output of Ottr as a string
        
        Returns:
            ``GradingResults``: the Ottr grading results
        """
        test_file_results = json.loads(ottr_output)["test_file_results"]
        test_files = []
        for tfr in test_file_results:
            test_cases, test_case_results = [], []

            for tcr in tfr["test_case_results"]:
                tc = tcr["test_case"]
                test_cases.append(TestCase(
                    name = tc["name"],
                    body = tc["code"],
                    hidden = tc["hidden"],
                    points = tc["points"],
                    success_message = tc["success_message"],
                    failure_message = tc["failure_message"],
                ))
                test_case_results.append(TestCaseResult(
                    test_case = test_cases[-1],
                    message = tcr["error"],
                    passed = tcr["passed"],
                ))
            
            test_file = OttrTestFile(
                name = os.path.splitext(os.path.basename(tfr["filename"]))[0],
                path = tfr["filename"],
                test_cases = test_cases,
            )
            test_file.test_case_results = test_case_results
            
            test_files.append(test_file)
        
        return cls(test_files)

    @property
    def test_files(self):
        """
        ``list[TestFile]``: the names of all test files tracked in these grading results
        """
        return list(self.results.keys())

    @property
    def total(self):
        """
        ``int`` or ``float``: the total points earned
        """
        return sum(tr.score for tr in self.results.values())

    @property
    def possible(self):
        """
        ``int`` or ``float``: the total points possible
        """
        return sum(tr.possible for tr in self.results.values())
    
    def get_result(self, test_name):
        """
        Returns the ``TestFile`` corresponding to the test with name ``test_name``

        Args:
            test_name (``str``): the name of the desired test
        
        Returns:
            ``TestFile``: the graded test file object
        """
        return self.results[test_name]

    def get_score(self, test_name):
        """
        Returns the score of a test tracked by these results

        Args:
            test_name (``str``): the name of the test
        
        Returns:
            ``int`` or ``float``: the score
        """
        result = self.results[test_name]
        return result.score

    def update_score(self, test_name, new_score):
        """
        Updates the values in the ``GradingTestCaseResult`` object stored in ``self.results[test_name]`` 
        with the key-value pairs in ``kwargs``.

        Args:
            test_name (``str``): the name of the test
            new_score (``int`` or ``float``): the new score
        """
        self.results[test_name].update_score(new_score)

    def set_output(self, output):
        """
        Updates the ``output`` field of the results JSON with text relevant to the entire submission.
        See https://gradescope-autograders.readthedocs.io/en/latest/specs/ for more information.

        Args:
            output (``str``): the output text
        """
        self.output = output

    def clear_results(self):
        """
        Empties the dictionary of results
        """
        self.results = {}

    def hide_everything(self):
        """
        Indicates that all results should be hidden from students on Gradescope
        """
        self.all_hidden = True
    
    def set_plugin_data(self, plugin_name, data):
        """
        Stores plugin data for plugin ``plugin_name`` in the results. ``data`` must be picklable.

        Args:
            plugin_name (``str``): the importable name of a plugin
            data (any): the data to store; must be serializable with ``pickle``
        """
        try:
            pickle.dumps(data)
        except:
            raise ValueError(f"Data was not pickleable: {data}")
        self._plugin_data[plugin_name] = data
    
    def get_plugin_data(self, plugin_name, default=None):
        """
        Retrieves data for plugin ``plugin_name`` in the results

        This method uses ``dict.get`` to retrive the data, so a ``KeyError`` is never raised if
        ``plugin_name`` is not found; rather, it returns ``None``.

        Args:
            plugin_name (``str``): the importable name of a plugin
            default (any, optional): a default value to return if ``plugin_name`` is not found

        Returns:
            any: the data stored for ``plugin_name`` if found
        """
        return self._plugin_data.get(plugin_name, default)

    def verify_against_log(self, log, ignore_hidden=True):
        """
        Verifies these scores against the results stored in this log using the results returned by 
        ``Log.get_results`` for comparison. Prints a message if the scores differ by more than the 
        default tolerance of ``math.isclose``. If ``ignore_hidden`` is ``True``, hidden tests are
        ignored when verifying scores.

        Args:
            log (``otter.check.logs.Log``): the log to verify against
            ignore_hidden  (``bool``, optional): whether to ignore hidden tests during verification

        Returns:
            ``bool``: whether a discrepancy was found
        """
        found_discrepancy = False
        # for test_name in  self.test_cases:
        for test_name, test_file in self.results.items():
            if ignore_hidden:
                tcrs = [test_case_result.passed for test_case_result in test_file.test_case_results if not test_case_result.hidden]
                score = sum(tcr.test_case.points for tcr in tcrs)
            else:
                score = test_file.score
            try:
                result = log.get_results(test_name)
                # TODO fix
                logged_score = result.score
                if not math.isclose(score, logged_score):
                    print("Score for {} ({:.3f}) differs from logged score ({:.3f})".format(
                        test_name, score, logged_score
                    ))
                    found_discrepancy = True
            except QuestionNotInLogException:
                print(f"No score for {test_name} found in this log")
                found_discrepancy = True
        return found_discrepancy

    def to_report_str(self):
        """
        Returns these results as a report string generated using the ``__repr__`` of the ``TestFile``
        class.

        Returns:
            ``str``: the report
        """
        return "\n".join(repr(test_file) for test_file in self.test_files)

    def to_dict(self):
        """
        Converts these results into a dictinary, extending the fields of the named tuples in ``results``
        into key, value pairs in a ``dict``.

        Returns:
            ``dict``: the results in dictionary form
        """
        return {tn: tf.to_dict() for tn, tf in self.results.items()}

    def summary(self, public_only=False):
        """
        Generate a summary of these results and return it as a string.

        Args:
            public_only (``bool``, optional): whether only public test cases should be included

        Returns:
            ``str``: the summary of results
        """
        return "\n\n".join(tf.summary(public_only=public_only) for _, tf in self.results.items())

    def to_gradescope_dict(self, config):
        """
        Converts these results into a dictionary formatted for Gradescope's autograder. Requires a 
        dictionary of configurations for the Gradescope assignment generated using Otter Generate.

        Args:
            config (``dict``): the grading configurations

        Returns:
            ``dict``: the results formatted for Gradescope
        """
        options = DEFAULT_OPTIONS.copy()
        options.update(config)

        output = {"tests": []}

        if self.output is not None:
            output["output"] = self.output
            # TODO: use output to display public test case results?

        # hidden visibility determined by show_hidden
        hidden_test_visibility = ("hidden", "after_published")[options["show_hidden"]]

        # if show_all_public is true and all tests are public tests, display all tests in results
        if options["show_all_public"] and all(tf.all_public for tf in self.results.values()):
            hidden_test_visibility = "visible"

        # start w/ summary of public tests
        output["tests"].append({
            "name": "Public Tests",
            "visibility": "visible",
            "output": self.summary(public_only=True),
        })

        for test_name in self.test_files:
            test_file = self.get_result(test_name)
            score, possible = test_file.score, test_file.possible

            output["tests"].append({
                "name": test_file.name,
                "score": score,
                "max_score": possible,
                "visibility": hidden_test_visibility,
                "output": test_file.summary(),
            })

        if options["show_stdout"]:
            output["stdout_visibility"] = "after_published"

        if options["points_possible"] is not None:
            try:
                output["score"] = self.total / self.possible * options["points_possible"]
            except ZeroDivisionError:
                output["score"] = 0

        if options["score_threshold"] is not None:
            try:
                if self.total / self.possible >= config["score_threshold"]:
                    output["score"] = options["points_possible"] or self.possible
                else:
                    output["score"] = 0
            except ZeroDivisionError:
                if 0 >= config["score_threshold"]:
                    output["score"] = options["points_possible"] or self.possible
                else:
                    output["score"] = 0

        if self.all_hidden:
            for test in output["tests"]:
                test["visibility"]  = "hidden"
            output["stdout_visibility"] = "hidden"
        
        return output



"""Runner classes for different languages"""

from .abstract_runner import AbstractLanguageRunner


def create_runner(otter_config, **kwargs):
    """
    Return an instantiated runner for the assignment based on user-specified configurations.
    """
    lang = AbstractLanguageRunner.determine_language(otter_config, **kwargs)

    if lang == "python":
        from .python_runner import PythonRunner
        return PythonRunner(otter_config, **kwargs)
    elif lang == "r":
        from .r_runner import RRunner
        return RRunner(otter_config, **kwargs)
    else:
        raise ValueError("Unsupported language: {}".format(lang))



import re

from .utils import Cell

from ..constants import TEST_REGEX
from ..r_adapter.tests import gen_suite
from ..utils import get_source

def is_test_cell(cell):
    """
    Returns whether the current cell is a test cell
    
    Args:
        cell (``otter.assign.rmarkdown_adapter.utils.Cell``): an Rmd file cell

    Returns:
        ``bool``: whether the cell is a test cell
    """
    if cell.cell_type != 'code':
        return False
    source = get_source(cell)
    return source and re.match(TEST_REGEX, source[1], flags=re.IGNORECASE)

def gen_test_cell(question, tests, tests_dict, assignment):
    """
    Parses a list of test named tuples and creates a single test file. Adds this test file as a value
    to ``tests_dict`` with a key corresponding to the test's name, taken from ``question``. Returns
    a code cell that runs the check on this test.
    
    Args:
        question (``dict``): question metadata
        tests (``list`` of ``Test``): tests to be written
        tests_dict (``dict``): the tests for this assignment
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``otter.assign.rmarkdown_adapter.utils.Cell``: code cell calling ``ottr::check`` on this test
    """
    source = f'```{{r}}\n. = ottr::check("tests/{question["name"]}.R")\n```'
    cell = Cell("code", source)

    points = question.get('points', len(tests))
    if isinstance(points, (int, float)):
        if points % len(tests) == 0:
            points = [points // len(tests) for _ in range(len(tests))]
        else:
            points = [points / len(tests) for _ in range(len(tests))]
    assert isinstance(points, list) and len(points) == len(tests), \
        f"Points for question {question['name']} could not be parsed:\n{points}"

    # update point values
    tests = [tc._replace(points=p) for tc, p in zip(tests, points)]
    test = gen_suite(question['name'], tests, points)

    tests_dict[question['name']] = test
    return cell



"""
Question configurations for Otter Assign
"""

import copy
import yaml

from .constants import BLOCK_QUOTE, ALLOWED_NAME
from .utils import get_source, lock, get_spec, EmptyCellException

def is_question_cell(cell):
    """
    Returns whether cell contains BEGIN QUESTION in a block quote
    
    Args:
        cell (``nbformat.NotebookNode``): a notebook cell
    
    Returns:
        ``bool``: whether the current cell is a question definition cell
    """
    if cell.cell_type != 'markdown':
        return False
    return get_spec(get_source(cell), "question") is not None

def gen_question_cell(cell, manual, need_close_export):
    """
    Returns a locked question cell with metadata hidden in an HTML comment
    
    Args:
        cell (``nbformat.NotebookNode``): the original question cell
    
    Returns:
        ``nbformat.NotebookNode``: the updated question cell
    """
    cell = copy.deepcopy(cell)
    source = get_source(cell)
    if manual:
        source = ["<!-- BEGIN QUESTION -->", ""] + source
    if need_close_export:
        source = ["<!-- END QUESTION -->", ""] + source
    begin_question_line = get_spec(source, "question")
    start = begin_question_line - 1
    assert source[start].strip() == BLOCK_QUOTE
    end = begin_question_line
    while source[end].strip() != BLOCK_QUOTE:
        end += 1
    source[start] = "<!--"
    source[end] = "-->"
    cell['source'] = '\n'.join(source)

    # check if cell is empty, and if so, throw error
    cell_text = source[:start]
    try:
        cell_text += source[end+1:]
    except IndexError:
        pass
    if not "".join(cell_text).strip():
        raise EmptyCellException()

    lock(cell)
    return cell

def read_question_metadata(cell):
    """
    Returns parsed question metadata from a question cell
    
    Args:
        cell (``nbformat.NotebookNode``): the question cell
    
    Returns:
        ``dict``: question metadata
    """
    source = get_source(cell)
    begin_question_line = get_spec(source, "question")
    i, lines = begin_question_line + 1, []
    while source[i].strip() != BLOCK_QUOTE:
        lines.append(source[i])
        i = i + 1
    metadata = yaml.full_load('\n'.join(lines))
    assert ALLOWED_NAME.match(metadata.get('name', '')), metadata
    return metadata



"""
Plugin for using Google Sheets to override scores for test cases
"""

import datetime as dt

from ..abstract_plugin import AbstractOtterPlugin


STRPTIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"


class RateLimiting(AbstractOtterPlugin):
    """
    Implements submission rate limiting by restricting the number of submissions a student is allowed
    during a specified window

    This plugin is configured by defining the duration of the window using the optional keys ``weeks``,
    ``days``, ``hours``, ``minutes``, ``seconds``, ``milliseconds``, and ``microseconds`` (all of which
    default to 0). The number of submissions allowed during this window is configured with 
    ``allowed_submissions``. For example, to allow only 2 submissions every hour-and-a-half:

    .. code-block:: json
    
        {
            "plugins": [
                {
                    "otter.plugins.builtin.RateLimiting": {
                        "hours": 1,
                        "minutes": 30,
                        "allowed_submissions": 2
                    }
                }
            ]
        }

    When a student submits, the window is caculated as a `datetime.timedelta` object and if the student 
    has at least `allowed_submissions` in that window, the submission's results are hidden and the 
    student is only shown a message; from the example above:

    .. code-block::
        
        You have exceeded the rate limit for the autograder. Students are allowed 2 submissions every 1 hours 30 minutes.

    The results of submission execution are still visible to instructors.

    If the student's submission is allowed based on the rate limit, the plugin outputs a message in 
    the plugin report; from the example above:

    .. code-block::

        Students are allowed 2 submissions every 1 hours 30 minutes. You have {number of previous submissions} submissions in that period.
    """

    def _window_to_str(self):
        """
        Returns a string representation of the configured window
        """
        result = ""
        for base in ["weeks", "days", "hours", "minutes", "seconds", "microseconds", "milliseconds"]:
            if self.plugin_config.get(base, 0):
                result += f"{self.plugin_config.get(base, 0)} {base} "
        return result.strip()

    def _submission_allowed(self):
        """
        Determines whether the rate limit was execeeded and returns a corresponding message

        Returns:
            ``tuple[bool,str]``: whether the submission is allowed and the message
        """
        window = dt.timedelta(
            self.plugin_config.get("days", 0),
            self.plugin_config.get("seconds", 0),
            self.plugin_config.get("microseconds", 0),
            self.plugin_config.get("milliseconds", 0),
            self.plugin_config.get("minutes", 0),
            self.plugin_config.get("hours", 0),
            self.plugin_config.get("weeks", 0),
        )

        subm_time = dt.datetime.strptime(self.submission_metadata["created_at"], STRPTIME_FORMAT)

        prev_subms = 0
        for subm in self.submission_metadata["previous_submissions"]:
            st = dt.datetime.strptime(subm["submission_time"], STRPTIME_FORMAT)
            if subm_time - st <= window:
                prev_subms += 1

        if prev_subms >= self.plugin_config["allowed_submissions"]:
            return False, \
                f"You have exceeded the rate limit for the autograder. Students are allowed {self.plugin_config['allowed_submissions']} " + \
                f"submissions every {self._window_to_str()}."

        else:
            return True, \
                f"Students are allowed {self.plugin_config['allowed_submissions']} submissions every {self._window_to_str()}. " + \
                f"You have {prev_subms} submissions in that period."

    def after_grading(self, results):
        """
        Determines whether a student has exceeded the rate limit; if yes, all results are hidden from
        the student using ``otter.test_files.GradingResults.hide_everything`` and the output of the
        grader is set to the message.

        Args:
            results (``otter.test_files.GradingResults.hide_everything``): the results of grading
        """
        if self.submission_metadata:
            allowed, output = self._submission_allowed()
            results.set_output(output)
            if not allowed:
                results.hide_everything()

    def generate_report(self):
        """
        Returns the message regarding whether the rate limit was exceeded for inclusion in the plugin
        report.

        Returns:
            ``str``: the message
        """
        if self.submission_metadata:
            return self._submission_allowed()[1]



"""Assignment block parsing for Otter Assign"""

import re
import yaml

from enum import Enum

from .utils import get_source, is_cell_type


class BlockType(Enum):
    """
    An enum of allowed block types.
    """

    QUESTION = "question"
    PROMPT = "prompt"
    SOLUTION = "solution"
    TESTS = "tests"


def is_block_boundary_cell(cell, block_type, end=False):
    """
    Determine whether ``cell`` is a boundary cell for a ``block_type`` block. If ``end`` is true,
    the block should be an end block; otherwise, it should be a begin block.

    Args:
        cell (``nbformat.NotebookNode``): the cell to check
        block_type (``BlockType``): the block type to check for
        end (``bool``, optional): whether to check for an end boundary instead of a begin

    Returns:
        ``bool``: whether the cell is a boundary cell of type ``block_type``
    """
    begin_or_end = 'end' if end else 'begin'
    regex = fr"#\s+{ begin_or_end }\s+{ block_type.value }\s*"
    source = get_source(cell)
    return is_cell_type(cell, "raw") and bool(re.match(regex, source[0], flags=re.IGNORECASE))


def is_assignment_config_cell(cell):
    """
    Determine whether ``cell`` is an assignment configuration cell.

    An assignment configuration cell is a raw cell starting with the line ``# ASSIGNMENT CONFIG``,
    e.g.

    .. code-block:: yaml

        # ASSIGNMENT CONFIG
        requirements: requirements.txt
        files:
            - data.csv

    Args:
        cell (``nbformat.NotebookNode``): the cell to check

    Returns:
        ``bool``: whether the cell is an assignment config cell
    """
    regex = r"#\s+assignment\s+config\s*"
    source = get_source(cell)
    return is_cell_type(cell, "raw") and bool(re.match(regex, source[0], flags=re.IGNORECASE))


def get_cell_config(cell):
    """
    Parse a cell's contents as YAML and return the resulting dictionary.

    Args:
        cell (``nbformat.NotebookNode``): the cell to check

    Returns:
        ``dict[str, object]``: the parsed configurations

    Raises:
        ``TypeError``: if parsing the YAML does not return a dictionary
    """
    source = get_source(cell)
    config = yaml.full_load("\n".join(source))
    if not isinstance(config, dict) and config is not None:
        # TODO: make this error nicer?
        raise TypeError(f"Found a begin cell configuration that is not a dictionary: {cell}")
    return config



"""
Rmarkdown adapter for Otter Assign
"""

import os
import copy
import pathlib
import nbformat

from .solutions import is_markdown_solution_cell
from .tests import gen_test_cell, is_test_cell
from .utils import rmd_to_cells, Cell, collapse_empty_cells

from ..assignment import is_assignment_cell, read_assignment_metadata
from ..questions import is_question_cell, read_question_metadata
from ..r_adapter.tests import read_test
from ..solutions import has_seed
from ..tests import any_public_tests
from ..utils import is_ignore_cell, EmptyCellException

def transform_notebook(rmd_string, assignment):
    """
    Converts a master RMd file to an Ottr-formatted solutions Rmd file, parsing test cells into
    dictionaries ready to be written as test files.

    Args:
        rmd_string (``str``): the Rmd document as a string
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``tuple[str, dict]``: the transformed Rmd file and a dictionary mapping test names to their 
            parsed contents

    """
    cells = rmd_to_cells(rmd_string)
    transformed_cells, test_files = get_transformed_cells(cells, assignment)
    collapse_empty_cells(transformed_cells)

    transformed_rmd_string = "\n".join([cell.source for cell in transformed_cells])

    return transformed_rmd_string, test_files

def get_transformed_cells(cells, assignment):
    """
    Takes in a list of cells from the master notebook and returns a list of cells for the solutions
    notebook. Replaces test cells with a cell calling ``ottr::check``, inserts Markdown response cells 
    for manual questions with Markdown solutions, and removes question metadata in question cells, 
    among other things.

    Args:
        cells (``list`` of ``otter.assign.rmarkdown_adapter.utils.Cell``): original code cells
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
    
    Returns:
        ``tuple[list, dict]``: list of cleaned notebook cells and a dictionary mapping test names to 
            their parsed contents
    """
    transformed_cells, test_files = [], {}
    question_metadata, test_cases, processed_solution, md_has_prompt = {}, [], False, False
    need_close_export, no_solution, in_collapse_block = False, False, False

    for cell in cells:

        if not cell.source.strip():
            transformed_cells.append(cell)
            continue

        if has_seed(cell):
            assignment.seed_required = True            

        if is_ignore_cell(cell):
            continue

        # this is the prompt cell or if a manual question then the solution cell
        if question_metadata and not processed_solution:
            assert not is_question_cell(cell), f"Found question cell before end of previous question cell: {cell}"

            if is_test_cell(cell):
                no_solution = True
                test = read_test(cell, question_metadata, assignment, rmd=True)
                test_cases.append(test)

            if not no_solution:
                transformed_cells.append(cell)
            
            processed_solution = True

        # if this is a test cell, parse and add to test_cases
        elif question_metadata and processed_solution and is_test_cell(cell):
            test = read_test(cell, question_metadata, assignment, rmd=True)
            test_cases.append(test)

        else:
            # the question is over -- we've seen the question and solution and any tests and now we 
            # need to get ready to process the next question which *could be this cell*
            if question_metadata and processed_solution:

                # create a Notebook.check cell
                if test_cases:
                    check_cell = gen_test_cell(question_metadata, test_cases, test_files, assignment)

                    # only add to notebook if there's a response cell or if there are public tests
                    if not no_solution or any_public_tests(test_cases):
                        transformed_cells.append(check_cell)
                
                # reset vars
                question_metadata, processed_solution, test_cases, md_has_prompt, no_solution = {}, False, [], False, False

            # update assignment config if present; don't add cell to output nb
            if is_assignment_cell(cell):
                assignment.update(read_assignment_metadata(cell))

            # if a question cell, parse metadata, comment out question metadata, and append to nb
            elif is_question_cell(cell):
                question_metadata = read_question_metadata(cell)
                manual = question_metadata.get('manual', False)

            elif is_markdown_solution_cell(cell):
                # transformed_cells.append(gen_markdown_response_cell())
                transformed_cells.append(cell)

            else:
                assert not is_test_cell(cell), f"Test outside of a question: {cell}"

                transformed_cells.append(cell)

    if test_cases:
        check_cell = gen_test_cell(question_metadata, test_cases, test_files, assignment)
        if not no_solution or any_public_tests(test_cases):
            transformed_cells.append(check_cell)

    return transformed_cells, test_files



"""PDF via HTML exporter"""

import nbconvert
import os
import pdfkit
import pkg_resources
import shutil
import tempfile

from contextlib import redirect_stdout, redirect_stderr
from io import StringIO, BytesIO
from nbconvert.exporters import export
from PyPDF2 import PdfFileMerger

from .base_exporter import BaseExporter, NBCONVERT_6, TEMPLATE_DIR
from .utils import notebook_pdf_generator


class PDFViaHTMLExporter(BaseExporter):
    """
    Exports notebooks to PDF files using HTML as an intermediary

    Converts IPython notebooks to PDFs by first converting them into temporary HTML files that are then
    converted to PDFs using wkhtmltopdf and its Python API pdfkit which are then stitched together (if
    pagebreaks are enabled) using PyPDF2.

    Attributes:
        default_options (``dict``): the default options for this exporter
    """

    default_options = BaseExporter.default_options.copy()
    default_options.update({
        "save_html": False,
        "template": "via_html"
    })

    @classmethod
    def convert_notebook(cls, nb_path, dest, **kwargs):
        assert shutil.which("wkhtmltopdf") is not None, "Cannot export via HTML without wkhtmltopdf"

        options = cls.default_options.copy()
        options.update(kwargs)

        nb = cls.load_notebook(nb_path, filtering=options["filtering"], pagebreaks=options["pagebreaks"])

        if NBCONVERT_6:
            nbconvert.TemplateExporter.extra_template_basedirs = [TEMPLATE_DIR]
            orig_template_name = nbconvert.TemplateExporter.template_name
            nbconvert.TemplateExporter.template_name = options["template"]

        exporter = nbconvert.HTMLExporter()
        if not NBCONVERT_6:
            exporter.template_file = os.path.join(TEMPLATE_DIR, options["template"] + ".tpl")

        if options["save_html"]:
            html, _ = export(exporter, nb)
            html_path = os.path.splitext(dest)[0] + ".html"
            with open(html_path, "wb+") as f:
                f.write(html.encode("utf-8"))
        
        merger = PdfFileMerger()
        for subnb in notebook_pdf_generator(nb):
            html, _ = export(exporter, subnb)

            pdfkit_options = {
                'enable-local-file-access': None, 
                'quiet': '', 
                'print-media-type': '', 
                'javascript-delay': 2000
            }
            pdf_contents = pdfkit.from_string(html, False, options=pdfkit_options)

            output = BytesIO()
            output.write(pdf_contents)
            output.seek(0)

            merger.append(output, import_bookmarks=False)

        merger.write(dest)

        if NBCONVERT_6:
            nbconvert.TemplateExporter.template_name = orig_template_name



"""Execution of a Jupyter Notebook"""

import os
import ast
import copy

from contextlib import redirect_stdout, redirect_stderr
from IPython.display import display
from unittest import mock

try:
    from IPython.core.inputtransformer2 import TransformerManager
    _IPYTHON_7 = True
except ImportError:
    from IPython.core.inputsplitter import IPythonInputSplitter
    _IPYTHON_7 = False

from .check_wrapper import CheckCallWrapper
from .transforms import create_collected_check_cell

from ..utils import id_generator


def execute_notebook(nb, check_results_list_name="check_results_secret", initial_env=None, 
                     ignore_errors=False, cwd=None, test_dir=None, seed=None, seed_variable=None):
    """
    Execute a notebook and return the global environment that results from execution.

    If ``ignore_errors`` is true, exceptions are swallowed.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to execute
        check_results_list_name (``str``, optional): the name of the list to collect check results in
        initial_env (``dict``, optional): an initial environment
        ignore_errors (``bool``, optional): whether exceptions should be ignored
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        seed (``int``, optional): random seed for intercell seeding
        seed_variable (``str``, optional): a variable name to override with the seed

    Results:
        ``dict``: global environment resulting from executing all code of the input notebook
    """
    if initial_env:
        global_env = initial_env.copy()
    else:
        global_env = {}

    # add display from IPython
    global_env["display"] = display

    from ..check.notebook import Notebook
    Notebook._tests_dir_override = test_dir if test_dir is not None else './tests'

    # add dummy Notebook class so that we can collect results w/out altering how the CheckCallWrapper
    # needs to function
    secret = id_generator()
    notebook_class_name = f"Notebook_{secret}"
    global_env[notebook_class_name] = Notebook

    source = ""

    if cwd:
        source = f"import sys\nsys.path.append(r\"{cwd}\")\n"
        exec(source, global_env)
    
    if seed is not None and seed_variable is None:
        import numpy as np
        import random
        global_env["np"] = np
        global_env["random"] = random

    if test_dir is None:
        test_dir = "/home/tests"

    for cell in nb['cells']:
        if cell['cell_type'] == 'code':
            if _IPYTHON_7:
                isp = TransformerManager()
            else:
                isp = IPythonInputSplitter(line_input_checker=False)

            try:
                code_lines = []
                cell_source_lines = cell['source']
                source_is_str_bool = False
                if isinstance(cell_source_lines, str):
                    source_is_str_bool = True
                    cell_source_lines = cell_source_lines.split('\n')

                for line in cell_source_lines:
                    if not line.startswith('%') and  "interact(" not in line:
                        code_lines.append(line)
                        if source_is_str_bool:
                            code_lines.append('\n')

                if seed is not None:
                    if seed_variable is None:
                        cell_source = f"np.random.seed({seed})\nrandom.seed({seed})\n" + \
                            isp.transform_cell(''.join(code_lines))
                    else:
                        cell_source = f"{seed_variable} = {seed}\n" + \
                            isp.transform_cell(''.join(code_lines))

                else:
                    cell_source = isp.transform_cell(''.join(code_lines))

                # patch otter.Notebook.export so that we don't create PDFs in notebooks
                # TODO: move this patch into CheckCallWrapper
                m = mock.mock_open()
                with mock.patch('otter.Notebook.export', m), mock.patch("otter.Notebook._log_event", m):
                    exec(cell_source, global_env)
                source += cell_source

            except:
                if not ignore_errors:
                    raise

        source += create_collected_check_cell(
            cell, check_results_list_name, notebook_class_name, test_dir)

    tree = ast.parse(source)
    transformer = CheckCallWrapper(check_results_list_name)
    tree = transformer.visit(tree)
    ast.fix_missing_locations(tree)

    try:
        cleaned_source = compile(tree, filename="nb-ast", mode="exec")
        with open(os.devnull, 'w') as f, redirect_stdout(f), redirect_stderr(f):
            # patch otter.Notebook.export so that we don't create PDFs in notebooks
            m = mock.mock_open()
            with mock.patch("otter.Notebook.export", m), mock.patch("otter.Notebook._log_event", m):
                exec(cleaned_source, global_env)

    except:
        if not ignore_errors:
            raise
    
    # reset tests dir
    Notebook._tests_dir_override = None

    return global_env



"""Utilities for Otter Assign"""

import copy
import datetime as dt
import json
import os
import pathlib
import pprint
import re
import shutil

from contextlib import contextmanager
from glob import glob
from textwrap import indent

from .constants import SEED_REGEX, BLOCK_QUOTE, IGNORE_REGEX

from ..execute import grade_notebook
from ..generate import main as generate_autograder
from ..generate.token import APIClient
from ..utils import get_relpath, get_source


class EmptyCellException(Exception):
    """
    Exception for empty cells to indicate deletion
    """


class AssignNotebookFormatException(Exception):
    """
    """
    def __init__(self, message, question_metadata, cell_index, *args, **kwargs):
        question_name = question_metadata.get("name")
        message += " ("
        if question_name is not None:
            message += f"question { question_name }, "
        message = message + f"cell number { cell_index + 1 })"
        super().__init__(message, *args, **kwargs)


#---------------------------------------------------------------------------------------------------
# Getters
#---------------------------------------------------------------------------------------------------

def get_spec(source, begin):
    """
    Returns the line number of the spec begin line or ``None``. Converts ``begin`` to an uppercase 
    string and looks for a line matching ``f"BEGIN {begin.upper()}"``. Used for finding question and
    assignment metadata, which match ``BEGIN QUESTION`` and ``BEGIN ASSIGNMENT``, resp.
    
    Args:
        source (``list`` of ``str``): cell source as a list of lines of text
        begin (``str``): the spec to look for
    
    Returns:
        ``int``: line number of BEGIN ASSIGNMENT, if present
        ``None``: if BEGIN ASSIGNMENT not present in the cell
    """
    block_quotes = [
        i for i, line in enumerate(source) if line[:3] == BLOCK_QUOTE
    ]
    assert len(block_quotes) % 2 == 0, f"wrong number of block quote delimieters in {source}"

    begins = [
        block_quotes[i] + 1 for i in range(0, len(block_quotes), 2) 
        if source[block_quotes[i]+1].strip(' ') == f"BEGIN {begin.upper()}"
    ]
    assert len(begins) <= 1, f'multiple BEGIN {begin.upper()} blocks defined in {source}'
    
    return begins[0] if begins else None

def get_notebook_language(nb):
    """
    Parse the notebook kernel language and return it as a string.

    Args:
        nb (``nbformat.NotebookNode``): the notebook

    Returns:
        ``str``: the name of the language as a lowercased string
    """
    return nb["metadata"]["kernelspec"]["language"].lower()


#---------------------------------------------------------------------------------------------------
# Cell Type Checkers
#---------------------------------------------------------------------------------------------------

def is_ignore_cell(cell):
    """
    Returns whether the current cell should be ignored
    
    Args:
        cell (``nbformat.NotebookNode``): a notebook cell

    Returns:
        ``bool``: whether the cell is a ignored
    """
    source = get_source(cell)
    return source and re.match(IGNORE_REGEX, source[0], flags=re.IGNORECASE)

def is_cell_type(cell, cell_type):
    return cell["cell_type"] == cell_type


#---------------------------------------------------------------------------------------------------
# Cell Reformatters
#---------------------------------------------------------------------------------------------------

def remove_output(nb):
    """
    Removes all outputs from a notebook in-place
    
    Args:
        nb (``nbformat.NotebookNode``): a notebook
    """
    for cell in nb['cells']:
        if 'outputs' in cell:
            cell['outputs'] = []
        if 'execution_count' in cell:
            cell['execution_count'] = None

def remove_cell_ids(nb):
    """
    Removes all cell IDs from a notebook in-place
    
    Args:
        nb (``nbformat.NotebookNode``): a notebook
    """
    for cell in nb['cells']:
        if 'id' in cell:
            cell.pop('id')

def lock(cell):
    """
    Makes a cell non-editable and non-deletable in-place

    Args:
        cell (``nbformat.NotebookNode``): cell to be locked
    """
    m = cell['metadata']
    m["editable"] = False
    m["deletable"] = False

def add_tag(cell, tag):
    """
    """
    cell = copy.deepcopy(cell)
    if "tags" not in cell["metadata"]:
        cell["metadata"]["tags"] = []
    cell["metadata"]["tags"].append(tag)
    return cell

def has_tag(cell, tag):
    """
    """
    return tag in cell["metadata"].get("tags", [])

def remove_tag(cell, tag):
    """
    """
    cell = copy.deepcopy(cell)
    if "tags" not in cell["metadata"] or tag not in cell["metadata"]["tags"]:
        return cell
    cell["metadata"]["tags"].remove(tag)
    return cell


#---------------------------------------------------------------------------------------------------
# Miscellaneous
#---------------------------------------------------------------------------------------------------

def str_to_doctest(code_lines, lines):
    """
    Converts a list of lines of Python code ``code_lines`` to a list of doctest-formatted lines ``lines``

    Args:
        code_lines (``list``): list of lines of python code
        lines (``list``): set of characters used to create function name
    
    Returns:
        ``list`` of ``str``: doctest formatted list of lines
    """
    if len(code_lines) == 0:
        return lines
    line = code_lines.pop(0)
    if line.startswith(" ") or line.startswith("\t"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    elif bool(re.match(r"^except[\s\w]*:", line)) or line.startswith("elif ") or \
            line.startswith("else:") or line.startswith("finally:"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    elif len(lines) > 0 and lines[-1].strip().endswith("\\"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    else:
        return str_to_doctest(code_lines, lines + [">>> " + line])

def run_tests(nb_path, debug=False, seed=None, plugin_collection=None):
    """
    Runs tests in the autograder version of the notebook
    
    Args:
        nb_path (``pathlib.Path``): path to iPython notebooks
        debug (``bool``, optional): ``True`` if errors should not be ignored
        seed (``int``, optional): random seed for notebook execution
        plugin_collection(``otter.plugins.PluginCollection``, optional): plugins to run with tests
    """
    curr_dir = os.getcwd()
    os.chdir(nb_path.parent)

    results = grade_notebook(
        nb_path.name, tests_glob=glob(os.path.join("tests", "*.py")), cwd=os.getcwd(), 
    	test_dir=os.path.join(os.getcwd(), "tests"), ignore_errors = not debug, seed=seed,
        plugin_collection=plugin_collection
    )

    if results.total != results.possible:
        raise RuntimeError(f"Some autograder tests failed in the autograder notebook:\n" + \
            indent(results.summary(), '    '))

    os.chdir(curr_dir)

def write_otter_config_file(master, result, assignment):
    """
    Creates an Otter configuration file (a ``.otter`` file) for students to use Otter tools, including
    saving environments and submitting to an Otter Service deployment, using assignment configurations.
    Writes the resulting file to the ``autograder`` and ``student`` subdirectories of ``result``.

    Args:
        master (``pathlib.Path``): path to master notebook
        result (``pathlib.Path``): path to result directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
    """
    config = {}

    config["notebook"] = master.name
    config["save_environment"] = assignment.save_environment
    config["ignore_modules"] = assignment.ignore_modules

    if assignment.variables:
        config["variables"] = assignment.variables

    config_name = master.stem + '.otter'
    with open(result / 'autograder' / config_name, "w+") as f:
        json.dump(config, f, indent=4)
    with open(result / 'student' / config_name, "w+") as f:
        json.dump(config, f, indent=4)

# TODO: update for new assign format
def run_generate_autograder(result, assignment, gs_username, gs_password, plugin_collection=None):
    """
    Runs Otter Generate on the autograder directory to generate a Gradescope zip file. Relies on 
    configurations in ``assignment.generate``.

    Args:
        result (``pathlib.Path``): the path to the result directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        gs_username (``str``): Gradescope username for token generation
        gs_password (``str``): Gradescope password for token generation
        plugin_collection (``otter.plugins.PluginCollection``, optional): a plugin collection to pass
            to Otter Generate
    """
    generate_args = assignment.generate
    if generate_args is True:
        generate_args = {}

    if "pdfs" in generate_args:
        raise ValueError("The 'pdfs' key of 'generate' is no longer supported. Put any " + \
                         "'pdfs' configurations inside the 'generate' key itself.")

    if assignment.is_r:
        generate_args["lang"] = "r"

    curr_dir = os.getcwd()
    os.chdir(str(result / 'autograder'))

    # use temp tests dir
    test_dir = "tests"
    if assignment.is_python and not assignment.test_files and assignment._temp_test_dir is None:
        raise RuntimeError("Failed to create temp tests directory for Otter Generate")
    elif assignment.is_python and not assignment.test_files:
        test_dir = str(assignment._temp_test_dir)

    requirements = None
    if assignment.requirements:
        if assignment.is_r:
            requirements = 'requirements.R'
        else:
            requirements = 'requirements.txt'

    files = []
    if assignment.files:
        files += assignment.files

    if assignment.autograder_files:
        ag_dir = os.getcwd()
        os.chdir(curr_dir)
        output_dir  = result / 'autograder'

        # copy files
        for file in assignment.autograder_files:

            # if a directory, copy the entire dir
            if os.path.isdir(file):
                shutil.copytree(file, str(output_dir / os.path.basename(file)))

            else:
                # check that file is in subdir
                assert os.getcwd() in os.path.abspath(file), \
                    f"{file} is not in a subdirectory of the master notebook directory"
                file_path = pathlib.Path(file).resolve()
                rel_path = file_path.parent.relative_to(pathlib.Path(os.getcwd()))
                os.makedirs(output_dir / rel_path, exist_ok=True)
                shutil.copy(file, str(output_dir / rel_path))

        os.chdir(ag_dir)

        files += assignment.autograder_files

    # TODO: move this config out of the assignment metadata and into the generate key
    if assignment.variables:
        generate_args['serialized_variables'] = str(assignment.variables)

    if generate_args:
        with open("otter_config.json", "w+") as f:
            json.dump(generate_args, f, indent=2)

    # TODO: change generate_autograder so that only necessary kwargs are needed
    timestamp = dt.datetime.now().strftime("%Y_%m_%dT%H_%M_%S_%f")
    notebook_name = assignment.master.stem
    output_path = f"{notebook_name}-autograder_{timestamp}.zip"
    generate_autograder(
        tests_dir=test_dir,
        output_path=output_path,
        config="otter_config.json" if generate_args else None,
        lang="python" if assignment.is_python else "r",
        requirements=requirements,
        overwrite_requirements=assignment.overwrite_requirements,
        environment="environment.yml" if assignment.environment else None,
        no_environment=False,
        username=gs_username,
        password=gs_password,
        files=files,
        plugin_collection=plugin_collection,
        assignment=assignment,
    )

    # clean up temp tests dir
    if assignment._temp_test_dir is not None:
        shutil.rmtree(str(assignment._temp_test_dir))

    os.chdir(curr_dir)

@contextmanager
def patch_copytree():
    """
    A context manager patch for ``shutil.copytree` on WSL. Shamelessly stolen from https://bugs.python.org/issue38633
    (see for more information)
    """
    import errno, shutil
    orig_copyxattr = shutil._copyxattr
    
    def patched_copyxattr(src, dst, *, follow_symlinks=True):
        try:
            orig_copyxattr(src, dst, follow_symlinks=follow_symlinks)
        except OSError as ex:
            if ex.errno != errno.EACCES: raise
    
    shutil._copyxattr = patched_copyxattr

    yield

    shutil._copyxattr = orig_copyxattr



"""Argument parser for Otter command-line tools"""

import click

from .assign import main as assign
from .check import main as check
from .export import main as export
from .generate import main as generate
from .grade import _ALLOWED_EXTENSIONS
from .grade import main as grade
from .run import main as run
from .version import print_version_info


@click.group(invoke_without_command=True)
@click.option("--version", is_flag=True, help="Show the version and exit")
def cli(version):
    """
    Command-line utility for Otter-Grader, a Python-based autograder for Jupyter Notebooks, 
    RMarkdown files, and Python and R scripts. For more information, see 
    https://otter-grader.readthedocs.io/.
    """
    if version:
        print_version_info(logo=True)
        return


defaults = assign.__kwdefaults__
@cli.command("assign")
@click.argument("master", type=click.Path(exists=True, dir_okay=False))
@click.argument("result", type=click.Path())
@click.option("--no-run-tests", is_flag=True, help="Do not run the tests against the autograder notebook")
@click.option("--no-pdfs", is_flag=True, help="Do not generate PDFs; overrides assignment config")
@click.option("--username", help="Gradescope username for generating a token")
@click.option("--password", help="Gradescope password for generating a token")
@click.option("--debug", is_flag=True, help="Do not ignore errors in running tests for debugging")
@click.option("--v1", is_flag=True, help="Use Otter Assign format v1 instead of v0")
def assign_cli(*args, **kwargs):
    """
    Create distribution versions of the Otter Assign formatted notebook MASTER and write the
    results to the directory RESULT, which will be created if it does not already exist.
    """
    return assign(*args, **kwargs)


defaults = check.__kwdefaults__
@cli.command("check")
@click.argument("file", type=click.Path(exists=True, dir_okay=False))
@click.option("-q", "--question", help="A specific quetsion to grade")
@click.option("-t", "--tests-path", default=defaults["tests_path"], type=click.Path(exists=True, file_okay=False), help="Path to the direcotry of test files")
@click.option("--seed", type=click.INT, help="A random seed to be executed before each cell")
def check_cli(*args, **kwargs):
	"""
	Check the Python script or Jupyter Notebook FILE against tests.
	"""
	return check(*args, **kwargs)


defaults = export.__kwdefaults__
@cli.command("export")
@click.argument("src", type=click.Path(exists=True, dir_okay=False))
@click.argument("dest", required=False, type=click.Path())
@click.option("--filtering", is_flag=True, help="Whether the PDF should be filtered")
@click.option("--pagebreaks", is_flag=True, help="Whether the PDF should have pagebreaks between questions")
@click.option("-s", "--save", is_flag=True, help="Save intermediate file(s) as well")
@click.option("-e", "--exporter", default=defaults["exporter"], type=click.Choice(["latex", "html"]), help="Type of PDF exporter to use")
@click.option("--no-xecjk", is_flag=True, help="Force-disable xeCJK in Otter's LaTeX template")
def export_cli(*args, **kwargs):
    """
    Export a Jupyter Notebook SRC as a PDF at DEST with optional filtering.

    If unspecified, DEST is assumed to be the basename of SRC with a .pdf extension.
    """
    return export(*args, **kwargs)


defaults = generate.__kwdefaults__
@cli.command("generate")
@click.option("-t", "--tests-dir", default=defaults["tests_dir"], type=click.Path(exists=True, file_okay=False), help="Path to test files")
@click.option("-o", "--output-path", default=defaults["output_path"], type=click.Path(), help="Path at which to write autograder zip file")
@click.option("-c", "--config", type=click.Path(exists=True, dir_okay=False), help="Path to otter configuration file; ./otter_config.json automatically checked")
@click.option("--no-config", is_flag=True, help="Disable auto-inclusion of unspecified Otter config file at ./otter_config.json")
@click.option("-r", "--requirements", type=click.Path(exists=True, dir_okay=False), help="Path to requirements.txt file; ./requirements.txt automatically checked")
@click.option("--no-requirements", is_flag=True, help="Disable auto-inclusion of unespecified requirements file at ./requirements.txt")
@click.option("--overwrite-requirements", is_flag=True, help="Overwrite (rather than append to) default requirements for Gradescope; ignored if no REQUIREMENTS argument")
@click.option("-e", "--environment", type=click.Path(exists=True, dir_okay=False), help="Path to environment.yml file; ./environment.yml automatically checked (overwrite)")
@click.option("--no-environment", is_flag=True, help="Disable auto-inclusion of unespecified environment file at ./environment.yml")
@click.option("-l", "--lang", default=defaults["lang"], type=click.Choice(["python", "r"], case_sensitive=False), help="Assignment programming language; defaults to Python")
@click.option("--username", help="Gradescope username for generating a token")
@click.option("--password", help="Gradescope password for generating a token")
@click.option("--token", help="Gradescope token for uploading PDFs")
@click.argument("files", nargs=-1)
def generate_cli(*args, **kwargs):
    """
    Generate a zip file to configure an Otter autograder, including FILES as support files.
    """
    return generate(*args, **kwargs)


defaults = grade.__kwdefaults__
@cli.command("grade")

# necessary path arguments
@click.option("-p", "--path", default=defaults["path"], help="Path to directory of submissions")
@click.option("-a", "--autograder", default=defaults["autograder"], help="Path to autograder zip file")
@click.option("-o", "--output-dir", default=defaults["output_dir"], help="Directory to which to write output")

# submission format arguments
@click.option("-z", "--zips", is_flag=True, help="Whether submissions are zip files from Notebook.export")
@click.option("--ext", default=defaults["ext"], type=click.Choice(_ALLOWED_EXTENSIONS), help="The extension to glob for submissions")

# PDF export options
@click.option("--pdfs", is_flag=True, help="Whether to copy notebook PDFs out of containers")

# other settings and optional arguments
@click.option("-v", "--verbose", is_flag=True, help="Flag for verbose output")
@click.option("--containers", type=click.INT, help="Specify number of containers to run in parallel")
@click.option("--image", default=defaults["image"], help="Custom docker image to run on")
@click.option("--timeout", type=click.INT, help="Submission execution timeout in seconds")
@click.option("--no-network", is_flag=True, help="Disable networking in the containers")
@click.option("--no-kill", is_flag=True, help="Do not kill containers after grading")
@click.option("--debug", is_flag=True, help="Print stdout/stderr from grading for debugging")

@click.option("--prune", is_flag=True, help="Prune all of Otter's grading images")
@click.option("-f", "--force", is_flag=True, help="Force action (don't ask for confirmation)")
def grade_cli(*args, **kwargs):
    """
    Grade assignments locally using Docker containers.
    """
    return grade(*args, **kwargs)


defaults = run.__kwdefaults__
@cli.command("run")
@click.argument("submission")
@click.option("-a", "--autograder", default=defaults["autograder"], type=click.Path(exists=True, dir_okay=False), help="Path to autograder zip file")
@click.option("-o", "--output-dir", default=defaults["output_dir"], type=click.Path(exists=True, file_okay=False), help="Directory to which to write output")
@click.option("--no-logo", is_flag=True, help="Suppress Otter logo in stdout")
@click.option("--debug", is_flag=True, help="Do not ignore errors when running submission")
def run_cli(*args, **kwargs):
    """
    Run non-containerized Otter on a single submission.
    """
    return run(*args, **kwargs)



"""Abstract runner class for the autograder"""

import os
import shutil

from abc import ABC, abstractmethod

from ..constants import DEFAULT_OPTIONS


class AbstractLanguageRunner(ABC):
    """
    A class defining the logic of running the autograder and generating grading results.

    Args:
        otter_config (``dict[str:object]``): user-specified configurations to override the defaults
        **kwargs: other user-specified confirations to override the defaults

    Attributes:
        options (``dict[str:object]``): the grading options, including default values from 
            ``otter.run.run_autograder.constants.DEFAULT_OPTIONS``
    """

    def __init__(self, otter_config, **kwargs):
        self.options = DEFAULT_OPTIONS.copy()
        self.options.update(otter_config)
        self.options.update(kwargs)

    @staticmethod
    def determine_language(otter_config, **kwargs):
        """
        Determine the language of the assignment based on user-specified configurations.
        """
        return kwargs.get("lang", otter_config.get("lang", DEFAULT_OPTIONS["lang"]))

    def get_option(self, option):
        """
        Return the value of a configuration, including defaults.
        """
        return self.options[option]

    def get_options(self):
        """
        Return the options dictionary, including defaults.
        """
        return self.options

    def prepare_files(self):
        """
        Copies tests and support files needed for running the autograder.

        When this method is invoked, the working directory is assumed to already be 
        ``self.options["autograder_dir"]``.
        """
        # put files into submission directory
        if os.path.exists("./source/files"):
            for file in os.listdir("./source/files"):
                fp = os.path.join("./source/files", file)
                if os.path.isdir(fp):
                    if not os.path.exists(os.path.join("./submission", os.path.basename(fp))):
                        shutil.copytree(fp, os.path.join("./submission", os.path.basename(fp)))
                else:
                    shutil.copy(fp, "./submission")

        # copy the tests directory
        if os.path.exists("./submission/tests"):
            shutil.rmtree("./submission/tests")
        shutil.copytree("./source/tests", "./submission/tests")

    @abstractmethod
    def resolve_submission_path(self):
        """
        Determine the path to the submission file, performing any necessary transformations on the
        file.

        When this method is invoked, the working directory is assumed to already be 
        ``{self.options["autograder_dir"]}/submission``.
        """
        ...

    @abstractmethod
    def run(self):
        """
        Run the autograder according to the configurations in ``self.options``.

        When this method is invoked, the working directory is assumed to already be 
        ``self.options["autograder_dir"]``.

        Retuns:
            ``otter.test_files.GradingResults``: the results from grading the submission
        """
        ...



"""Autograder runner for Python assignments"""

import json
import os
import pickle
import shutil
import warnings
import zipfile

from glob import glob

from .abstract_runner import AbstractLanguageRunner
from ..utils import OtterRuntimeError
from ....check.logs import Log
from ....check.notebook import _OTTER_LOG_FILENAME
from ....execute import grade_notebook
from ....export import export_notebook
from ....generate.token import APIClient
from ....plugins import PluginCollection
from ....utils import chdir, print_full_width


class PythonRunner(AbstractLanguageRunner):

    def prepare_files(self):
        """
        Copies tests and support files needed for running the autograder.

        Also creates ``__init__.py`` files to allow for relative imports.
        """
        super().prepare_files()

        # create __init__.py files
        open("__init__.py", "a").close()
        open("submission/__init__.py", "a").close()

    def resolve_submission_path(self):
        nbs = glob("*.ipynb")

        if len(nbs) > 1:
            raise OtterRuntimeError("More than one IPYNB file found in submission")

        if len(nbs) == 1:
            subm_path = nbs[0]

        else:
            pys = glob("*.py")
            pys = list(filter(lambda f: f != "__init__.py", pys))
            if len(pys) > 1:
                raise OtterRuntimeError("More than one Python file found in submission")

            elif len(pys) == 1:
                subm_path = pys[0]

            else:
                raise OtterRuntimeError("No gradable files found in submission")

        return subm_path

    def write_pdf(self, nb_path):
        """
        Generate a PDF of a notebook at ``nb_path`` using the options in ``self.options`` and return
        the that to the PDF.
        """
        try:
            pdf_path = os.path.splitext(nb_path)[0] + ".pdf"
            export_notebook(
                nb_path, dest=pdf_path, filtering=self.options["filtering"], 
                pagebreaks=self.options["pagebreaks"], exporter_type="latex")

        except Exception as e:
            print(f"\n\nError encountered while generating and submitting PDF:\n{e}")

        return pdf_path

    def submit_pdf(self, client, pdf_path):
        """
        Upload a PDF to a Gradescope assignment for manual grading.

        Args:
            client (``otter.generate.token.APIClient``): the Gradescope client
            pdf_path (``str``): path to the PDF
        """
        try:
            # get student email
            with open("../submission_metadata.json", encoding="utf-8") as f:
                metadata = json.load(f)

            student_emails = []
            for user in metadata["users"]:
                student_emails.append(user["email"])

            for student_email in student_emails:
                client.upload_pdf_submission(
                    self.options["course_id"], self.options["assignment_id"], student_email, pdf_path)

            print("\n\nSuccessfully uploaded submissions for: {}".format(", ".join(student_emails)))

        except Exception as e:
            print(f"\n\nError encountered while generating and submitting PDF:\n{e}")

    def run(self):
        os.environ["PATH"] = f"{self.options['miniconda_path']}/bin:" + os.environ.get("PATH")

        with chdir("./submission"):

            subm_path = self.resolve_submission_path()

            # load plugins
            plugins = self.options["plugins"]

            if plugins:
                with open("../submission_metadata.json", encoding="utf-8") as f:
                    submission_metadata = json.load(f)

                plugin_collection = PluginCollection(
                    plugins, os.path.abspath(subm_path), submission_metadata)

            else:
                plugin_collection = None

            if plugin_collection:
                plugin_collection.run("before_grading", self.options)

            if self.options["token"] is not None:
                client = APIClient(token=self.options["token"])
                generate_pdf = True
                has_token = True

            else:
                generate_pdf = self.options["pdf"]
                has_token = False
                client = None

            if os.path.isfile(_OTTER_LOG_FILENAME):
                try:
                    log = Log.from_file(_OTTER_LOG_FILENAME, ascending=False)

                except Exception as e:
                    if self.options["grade_from_log"]:
                        raise e

                    else:
                        print(f"Could not deserialize the log due to an error:\n{e}")
                        log = None

            else:
                if self.options["grade_from_log"]:
                    raise OtterRuntimeError("Grade from log indicated but log not found")

                log = None

            scores = grade_notebook(
                subm_path, 
                tests_glob = glob("./tests/*.py"), 
                name = "submission", 
                cwd = os.getcwd(), 
                test_dir = "./tests",
                ignore_errors = not self.options["debug"], 
                seed = self.options["seed"],
                seed_variable = self.options["seed_variable"],
                log = log if self.options["grade_from_log"] else None,
                variables = self.options["serialized_variables"],
                plugin_collection = plugin_collection,
                script = os.path.splitext(subm_path)[1] == ".py",
            )

            if self.options["print_summary"]:
                print("\n\n\n\n", end="")
                print_full_width("-", mid_text="GRADING SUMMARY")

            # verify the scores against the log
            if self.options["print_summary"]:
                print()
                if log is not None:
                    try:
                        found_discrepancy = scores.verify_against_log(log)
                        if not found_discrepancy and self.options["print_summary"]:
                            print("No discrepancies found while verifying scores against the log.")

                    except BaseException as e:
                        print(f"Error encountered while trying to verify scores with log:\n{e}")

                else:
                    print("No log found with which to verify student scores.")

            if generate_pdf:
                pdf_path = self.write_pdf(subm_path)

                if has_token:
                    self.submit_pdf(client, pdf_path)

            if plugin_collection:
                report = plugin_collection.generate_report()
                if report.strip():
                    print("\n\n" + report)        

        return scores



"""
Solution removal for Otter Assign
"""

import re

from .utils import rmd_to_cells, collapse_empty_cells

from ..constants import MD_RESPONSE_CELL_SOURCE
from ..utils import get_source

def is_markdown_solution_cell(cell):
    """
    Returns whether any line of the cell matches `<!-- BEGIN SOLUTION -->`
    
    Args:
        cell (``otter.assign.rmarkdown_adapter.utils.Cell``): notebook cell
    
    Returns:
        ``bool``: whether the current cell is a Markdown solution cell
    """
    if not cell.cell_type == 'markdown':
        return False
    source = get_source(cell)
    return source and any([re.match(r"\s*<!-- BEGIN SOLUTION -->\s*", l, flags=re.IGNORECASE) for l in source])

solution_assignment_regex = re.compile(r"(\s*[a-zA-Z0-9_. ]*(=|<-))(.*)[ ]?#[ ]?SOLUTION")
def solution_assignment_sub(match):
    """
    Substitutes the first matching group  with ` NULL # YOUR CODE HERE`
    """
    prefix = match.group(1)
    return prefix + ' NULL # YOUR CODE HERE'

solution_line_regex = re.compile(r"(\s*)([^#\n]+)[ ]?#[ ]?SOLUTION")
def solution_line_sub(match):
    """
    Substitutes the first matching group  with `# YOUR CODE HERE`
    """
    prefix = match.group(1)
    return prefix + '# YOUR CODE HERE'

begin_solution_regex = re.compile(r"(\s*)# BEGIN SOLUTION( NO PROMPT)?")
skip_suffixes = [
    '# SOLUTION NO PROMPT', '# BEGIN PROMPT', '# END PROMPT', '# SEED', '<!-- BEGIN PROMPT -->',
    '<!-- END PROMPT -->'
]

begin_md_solution_regex = re.compile(r"(\s*)<!--\s*BEGIN SOLUTION( NO PROMPT)?\s*-->")

SUBSTITUTIONS = [
    (solution_assignment_regex, solution_assignment_sub),
    (solution_line_regex, solution_line_sub),
]

# TODO: comments, docstrings
def replace_solutions(lines):
    """
    Replaces solutions in ``lines``
    
    Args:
        lines (``list`` of ``str``): solutions as a list of strings

    Returns:
        ``list`` of ``str``: stripped version of lines without solutions
    """
    stripped = []
    solution = False
    for line in lines:

        # ...
        if any(line.endswith(s) for s in skip_suffixes):
            continue

        # ...
        if solution and not (line.endswith('# END SOLUTION') or line.rstrip().endswith("<!-- END SOLUTION -->")):
            continue

        # ...
        if line.endswith('# END SOLUTION') or line.rstrip().endswith("<!-- END SOLUTION -->"):
            assert solution, f"END SOLUTION without BEGIN SOLUTION in {lines}"
            solution = False
            continue

        # ...
        begin_solution = begin_solution_regex.match(line)
        if begin_solution:
            assert not solution, f"Nested BEGIN SOLUTION in {lines}"
            solution = True
            if not begin_solution.group(2):
                line = begin_solution.group(1) + '# YOUR CODE HERE'
            else:
                continue
        
        begin_solution = begin_md_solution_regex.match(line)
        if begin_solution:
            assert not solution, f"Nested BEGIN SOLUTION in {lines}"
            solution = True
            if not begin_solution.group(2):
                line = begin_solution.group(1) + MD_RESPONSE_CELL_SOURCE
            else:
                continue
        
        for exp, sub in SUBSTITUTIONS:
            m = exp.match(line)
            if m:
                line = sub(m)
        
        stripped.append(line)
    
    assert not solution, f"BEGIN SOLUTION without END SOLUTION in {lines}"
    
    return stripped

def strip_solutions_and_output(rmd_string):
    """
    Writes a Rmd file with solutions stripped
    
    Args:
        rmd_string (``str``): the Rmd document as a string
    """
    md_solutions = []
    cells = rmd_to_cells(rmd_string)
    for i, cell in enumerate(cells):
        cells[i] = cell._replace(source='\n'.join(replace_solutions(get_source(cell))))
    collapse_empty_cells(cells)
    rmd_string = "\n".join([c.source for c in cells])    
    return rmd_string



"""
Otter Assign command-line utility
"""

import json
import os
import pathlib
import warnings

from .assignment import Assignment
from .utils import run_tests, write_otter_config_file, run_generate_autograder

from ...export import export_notebook
from ...export.exporters import WkhtmltopdfNotFoundError
from ...plugins import PluginCollection
from ...utils import get_relpath, block_print


def main(master, result, *, no_pdfs=False, no_run_tests=False, username=None, password=None, 
         debug=False):
    """
    Runs Otter Assign on a master notebook
    
    Args:
        master (``str``): path to master notebook
        result (``str``): path to result directory
        no_pdfs (``bool``): whether to ignore any configurations indicating PDF generation for this run
        no_run_tests (``bool``): prevents Otter tests from being automatically run on the solutions 
            notebook
        username (``str``): a username for Gradescope for generating a token
        password (``str``): a password for Gradescope for generating a token
        debug (``bool``): whether to run in debug mode (without ignoring errors during testing)
        **kwargs: ignored kwargs (a remnant of how the argument parser is built)
    """
    master, result = pathlib.Path(os.path.abspath(master)), pathlib.Path(os.path.abspath(result))
    print("Generating views...")

    assignment = Assignment()

    # # check language
    # if lang is not None:
    #     lang = lang.lower()
    #     assert lang in ["r", "python"], f"Language {lang} is not valid"
    #     assignment.lang = lang
    
    # TODO: update this condition
    if True:
        result = get_relpath(master.parent, result)
        orig_dir = os.getcwd()
        os.chdir(master.parent)
        # master = pathlib.Path(master.name)
    
    assignment.master, assignment.result = master, result
    # assignment.files = files

    if assignment.is_rmd:
        from .rmarkdown_adapter.output import write_output_directories
    else:
        from .output import write_output_directories

    try:
        output_nb_path = write_output_directories(master, result, assignment)

        # check that we have a seed if needed
        if assignment.seed_required:
            generate_args = assignment.generate
            if generate_args is True:
                generate_args = {'seed': None}
            assert not generate_args or generate_args.get('seed', None) is not None or \
                not assignment.is_python, "Seeding cell found but no seed provided"

        plugins = assignment.plugins
        if plugins:
            pc = PluginCollection(plugins, "", {})
            pc.run("during_assign", assignment)
            if assignment.generate is True:
                assignment.generate = {"plugins": []}
            if assignment.generate is not False:
                if not assignment.generate.get("plugins"):
                    assignment.generate["plugins"] = []
                assignment.generate["plugins"].extend(plugins)
        else:
            pc = None
        
        # generate PDF of solutions
        if assignment.solutions_pdf and not assignment.is_rmd and not no_pdfs:
            print("Generating solutions PDF...")
            filtering = assignment.solutions_pdf == 'filtered'

            try:
                export_notebook(
                    str(result / 'autograder' / master.name),
                    dest=str(result / 'autograder' / (master.stem + '-sol.pdf')),
                    filtering=filtering,
                    pagebreaks=filtering,
                    exporter_type="html",
                )
            except WkhtmltopdfNotFoundError:
                export_notebook(
                    str(result / 'autograder' / master.name),
                    dest=str(result / 'autograder' / (master.stem + '-sol.pdf')),
                    filtering=filtering,
                    pagebreaks=filtering,
                )

        # generate a tempalte PDF for Gradescope
        if not assignment.is_rmd and assignment.template_pdf and not no_pdfs:
            print("Generating template PDF...")
            export_notebook(
                str(result / 'autograder' / master.name),
                dest=str(result / 'autograder' / (master.stem + '-template.pdf')), 
                filtering=True, 
                pagebreaks=True, 
                exporter_type="latex",
            )

        # generate the .otter file if needed
        if not assignment.is_rmd and assignment.save_environment:
            if assignment.is_r:
                warnings.warn(
                    "Otter Service and serialized environments are unsupported with R, "
                    "configurations ignored"
                )
            else:
                write_otter_config_file(master, result, assignment)

        # generate Gradescope autograder zipfile
        if assignment.generate:
            # TODO: move this to another function
            print("Generating autograder zipfile...")
            run_generate_autograder(result, assignment, username, password, plugin_collection=pc)

        # run tests on autograder notebook
        if assignment.run_tests and not no_run_tests and assignment.is_python:
            print("Running tests...")
            # with block_print():
            if isinstance(assignment.generate, bool):
                seed = None
            else:
                seed = assignment.generate.get('seed', None)
            
            if assignment._otter_config is not None:
                test_pc = PluginCollection(assignment._otter_config.get("plugins", []), output_nb_path, {})

            else:
                test_pc = pc

            run_tests(result / 'autograder' / master.name, debug=debug, seed=seed, plugin_collection=test_pc)
            print("All tests passed!")
    
    # for tests
    except:
        # TODO: change this condition
        if True:
            os.chdir(orig_dir)

        raise

    else:
        # TODO: change this condition
        if True:
            os.chdir(orig_dir)



"""Docker container management for Otter Grade"""
import glob
import os
import pandas as pd
import pickle
import pkg_resources
import shutil
import tempfile
import zipfile

from concurrent.futures import ThreadPoolExecutor, wait
from python_on_whales import docker
from typing import Optional

from .utils import generate_hash, OTTER_DOCKER_IMAGE_TAG


def build_image(zip_path, base_image, tag):
    """
    Creates a grading image based on the autograder zip file and attaches a tag.

    Args:
        zip_path (``str``): path to the autograder zip file
        base_image (``str``): base Docker image to build from
        tag (``str``): tag to be added when creating the image

    Returns:
        ``str``: the tag of the newly-build Docker image
    """
    image = OTTER_DOCKER_IMAGE_TAG + ":" + tag
    dockerfile = pkg_resources.resource_filename(__name__, "Dockerfile")

    if not docker.image.exists(image):
        print(f"Building new image using {base_image} as base image")

        tmp_dir = tempfile.mkdtemp()
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            zip_ref.extractall(tmp_dir)

        docker.build(tmp_dir, build_args={
            "BASE_IMAGE": base_image
        }, tags=[image], file=dockerfile, load=True)
        shutil.rmtree(tmp_dir)
    return image


def launch_grade(zip_path, submissions_dir, verbose=False, num_containers=None, ext="ipynb", 
                 no_kill=False, output_path="./", debug=False, zips=False,
                 image="ucbdsinfra/otter-grader", pdfs=False, timeout=None, network=True):
    """
    Grades notebooks in parallel Docker containers

    This function runs ``num_containers`` Docker containers in parallel to grade the student submissions
    in ``submissions_dir`` using the autograder configuration file at ``zip_path``. It can additionally 
    generate PDFs for the parts of the assignment needing manual grading.

    Args:
        zip_path(``str``): path to zip file used to set up container
        submissions_dir (``str``): path to directory of student submissions to be graded
        verbose (``bool``, optional): whether status messages should be printed to the command line
        num_containers (``int``, optional): The number of parallel containers that will be run
        ext (``str``, optional): the submission file extension for globbing
        no_kill (``bool``, optional): whether the grading containers should be kept running after
            grading finishes
        output_path (``str``, optional): path at which to write grades CSVs copied from the container
        debug (``bool``, optional): whether to run grading in debug mode (prints grading STDOUT and STDERR
            from each container to the command line)
        zips (``bool``, optional): whether the submissions are zip files formatted from ``Notebook.export``
        image (``str``, optional): a base image to use for building Docker images
        pdfs (``bool``, optional): whether to copy PDFs out of the containers
        timeout (``int``): timeout in seconds for each container
        network (``bool``): whether to enable networking in the containers

    Returns:
        ``list`` of ``pandas.core.frame.DataFrame``: the grades returned by each container spawned during
            grading
    """
    if not num_containers:
        num_containers = 4

    pool = ThreadPoolExecutor(num_containers)
    futures = []
    img = build_image(zip_path, image, generate_hash(zip_path))

    if zips:
        pattern = "*.zip"
    else:
        pattern = f"*.{ext}"

    submissions = glob.glob(os.path.join(submissions_dir, pattern))
    pdf_dir = os.path.join(output_path, "submission_pdfs")

    for subm_path in submissions:
        futures += [
            pool.submit(
                grade_assignments,
                submission_path=subm_path,
                verbose=verbose,
                image=img,
                no_kill=no_kill,
                pdf_dir=pdf_dir,
                debug=debug,
                pdfs=pdfs,
                timeout=timeout,
                network=network,
            )
        ]

    # stop execution while containers are running
    finished_futures = wait(futures)

    # return list of dataframes
    return [df.result() for df in finished_futures[0]]


def grade_assignments(submission_path, image, verbose=False, no_kill=False, pdf_dir=None, 
                      debug=False, pdfs=False, timeout: Optional[int] = None, network=True):
    """
    Grades multiple submissions in a directory using a single docker container. If no PDF assignment is
    wanted, set all three PDF params (``unfiltered_pdfs``, ``tag_filter``, and ``html_filter``) to ``False``.

    Args:
        submission_path (``str``): path to the submission to be graded
        image (``str``): a Docker image tag to be used for grading environment
        verbose (``bool``, optional): whether status messages should be printed to the command line
        no_kill (``bool``, optional): whether the grading containers should be kept running after
            grading finishes
        pdf_dir (``str``, optional): directory in which to put notebook PDFs, if applicable
        debug (``bool``, False): whether to run grading in debug mode (prints grading STDOUT and STDERR
            from each container to the command line)
        pdfs (``bool``, optional): whether to copy PDFs out of the containers
        timeout (``int``): timeout in seconds for each container
        network (``bool``): whether to enable networking in the containers

    Returns:
        ``pandas.core.frame.DataFrame``: A dataframe of file to grades information
    """
    temp_subm_file, temp_subm_path = tempfile.mkstemp()
    shutil.copyfile(submission_path, temp_subm_path)

    results_file, results_path = tempfile.mkstemp(suffix=".pkl")
    pdf_path = None
    if pdfs:
        pdf_file, pdf_path = tempfile.mkstemp(suffix=".pdf")

    try:
        nb_basename = os.path.basename(submission_path)
        nb_name = os.path.splitext(nb_basename)[0]

        volumes = [
            (temp_subm_path, f"/autograder/submission/{nb_basename}"),
            (results_path, "/autograder/results/results.pkl")
        ]
        if pdfs:
            volumes.append((pdf_path, f"/autograder/submission/{nb_name}.pdf"))

        args = {}

        if network is not None and not network:
            args['networks'] = 'none'

        container = docker.container.run(image, command=["/autograder/run_autograder"], volumes=volumes, detach=True, **args)

        if timeout:
            import threading

            def kill_container():
                docker.container.kill(container)

            timer = threading.Timer(timeout, kill_container)
            timer.start()

        container_id = container.id[:12]
        if verbose:
            print(f"Grading {submission_path} in container {container_id}...")

        exit = docker.container.wait(container)

        if timeout:
            timer.cancel()

        if debug:
            print(docker.container.logs(container))

        if not no_kill:
            container.remove()

        if exit != 0:
            raise Exception(f"Executing '{submission_path}' in docker container failed! Exit code: {exit}")

        with open(results_path, "rb") as f:
            scores = pickle.load(f)

        scores = scores.to_dict()
        scores = {t: [scores[t]["score"]] if type(scores[t]) == dict else scores[t] for t in scores}
        scores["file"] = os.path.split(submission_path)[1]
        df = pd.DataFrame(scores)

        if pdfs:
            os.makedirs(pdf_dir, exist_ok=True)

            local_pdf_path = os.path.join(pdf_dir, f"{nb_name}.pdf")
            shutil.copy(pdf_path, local_pdf_path)

    finally:
        os.close(results_file)
        os.remove(results_path)
        os.close(temp_subm_file)
        os.remove(temp_subm_path)
        if pdfs:
            os.close(pdf_file)
            os.remove(pdf_path)

    return df



"""
ottr test adapters for Otter Assign
"""

import re
import pprint
import yaml
import nbformat

from collections import namedtuple

from ..constants import TEST_REGEX, OTTR_TEST_NAME_REGEX, OTTR_TEST_FILE_TEMPLATE
from ..tests import write_test
from ..utils import get_source, lock

Test = namedtuple('Test', ['name', 'hidden', 'points', 'body', 'success_message', 'failure_message'])

def read_test(cell, question, assignment, rmd=False):
    """
    Returns the contents of a test as a ``(name, hidden, body)`` named tuple
    
    Args:
        cell (``nbformat.NotebookNode``): a test cell
        question (``dict``): question metadata
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        rmd (``bool``, optional): whether the cell is from an Rmd file; if true, the first and last
            lines of ``cell``'s source are trimmed, since they should be backtick delimeters

    Returns:
        ``Test``: test named tuple
    """
    if rmd:
        source = get_source(cell)[1:-1]
    else:
        source = get_source(cell)
    hidden = bool(re.search("hidden", source[0], flags=re.IGNORECASE))
    lines = source[1:]
    test_name = None
    # for line in lines:
    #     match = re.match(OTTR_TEST_NAME_REGEX, line)
    #     if match:
    #         test_name = match.group(1)
    #         break
    # assert test_name is not None, f"Could not parse test name:\n{cell}"
    # TODO: hook up success_message and failure_message
    # TODO: add parsing for TEST CONFIG blocks
    return Test(test_name, hidden, None, '\n'.join(lines), "", "")

def gen_test_cell(question, tests, tests_dict, assignment):
    """
    Parses a list of test named tuples and creates a single test file. Adds this test file as a value
    to ``tests_dict`` with a key corresponding to the test's name, taken from ``question``. Returns
    a code cell that runs the check on this test.
    
    Args:
        question (``dict``): question metadata
        tests (``list`` of ``Test``): tests to be written
        tests_dict (``dict``): the tests for this assignment
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``nbformat.NotebookNode``: code cell calling ``ottr::check`` on this test
    """
    cell = nbformat.v4.new_code_cell()
    cell.source = ['. = ottr::check("tests/{}.R")'.format(question['name'])]

    points = question.get('points', len(tests))
    if isinstance(points, (int, float)):
        if points % len(tests) == 0:
            points = [points // len(tests) for _ in range(len(tests))]
        else:
            points = [points / len(tests) for _ in range(len(tests))]
    assert isinstance(points, list) and len(points) == len(tests), \
        f"Points for question {question['name']} could not be parsed:\n{points}"

    # update point values
    tests = [tc._replace(points=p) for tc, p in zip(tests, points)]
    test = gen_suite(question['name'], tests, points)

    tests_dict[question['name']] = test
    lock(cell)
    return cell

def gen_suite(name, tests, points):
    """
    Generates an R-formatted test file for ottr

    Args:
        name (``str``): the test name
        tests (``list`` of ``Test``): the test case named tuples that define this test file
        points (``float`` or ``int`` or ``list`` of ``float`` or ``int``): th points per question

    Returns:
        ``str``: the rendered R test file
    """
    template_data = {'name': name, 'test_cases': tests}
    return OTTR_TEST_FILE_TEMPLATE.render(**template_data)

def remove_hidden_tests_from_dir(nb, test_dir, assignment, use_files=True):
    """
    Rewrites test files in a directory to remove hidden tests
    
    Args:
        nb (``nbformat.NotebookNode``): the student notebook
        test_dir (``pathlib.Path``): path to test files directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        use_files (``bool``, optional): ignored for R assignments
    """
    for f in test_dir.iterdir():
        if f.suffix != '.R':
            continue

        with open(f) as f2:
            test = f2.read()
        
        test = re.sub(r"    ottr::TestCase\$new\(\s*hidden = TRUE[\w\W]+?^    \),?", "", test, flags=re.MULTILINE)
        test = re.sub(r",(\s*  \))", r"\1", test, flags=re.MULTILINE)  # removes a trailing comma if present

        write_test({}, f, test, use_file=True)



"""Assignment configurations for Otter Assign"""

import yaml

from .constants import BLOCK_QUOTE
from .utils import get_source, get_spec
from ..utils import convert_config_description_dict


_DEFAULT_ASSIGNMENT_CONFIGURATIONS_WITH_DESCRIPTIONS = [
    {
        "key": "requirements",
        "description": "the path to a requirements.txt file or a list of packages",
        "default": None,
    },
    {
        "key": "overwrite_requirements",
        "description": "whether to overwrite Otter's default requirement.txt in Otter Generate",
        "default": False,
    },
    {
        "key": "environment",
        "description": "the path to a conda environment.yml file",
        "default": None,
    },
    {
        "key": "run_tests",
        "description": "whether to run the assignment tests against the autograder notebook",
        "default": True,
    },
    {
        "key": "solutions_pdf",
        "description": "whether to generate a PDF of the solutions notebook",
        "default": False,
    },
    {
        "key": "template_pdf",
        "description": "whether to generate a filtered Gradescope assignment template PDF",
        "default": False,
    },
    {
        "key": "init_cell",
        "description": "whether to include an Otter initialization cell in the output notebooks",
        "default": True,
    },
    {
        "key": "check_all_cell",
        "description": "whether to include an Otter check-all cell in the output notebooks",
        "default": True,
    },
    {
        "key": "export_cell",
        "description": "whether to include an Otter export cell in the output notebooks",
        "default": [
            {
                "key": "instructions",
                "description": "additional submission instructions to include in the export cell",
                "default": "",
            },
            {
                "key": "pdf",
                "description": "whether to include a PDF of the notebook in the generated zip file",
                "default": True,
            },
            {
                "key": "filtering",
                "description": "whether the generated PDF should be filtered",
                "default": True,
            },
            {
                "key": "force_save",
                "description": "whether to force-save the notebook with JavaScript (only works in " \
                    "classic notebook)",
                "default": False,
            },
            {
                "key": "run_tests",
                "description": "whether to run student submissions against local tests during export",
                "default": False,
            },
        ],
    },
    {
        "key": "seed",
        "description": "intercell seeding configurations",
        "default": None,
        "subkeys": [
            {
                "key": "variable",
                "description": "a variable name to override with the autograder seed during grading",
                "default": None,
            },
            {
                "key": "autograder_value",
                "description": "the value of the autograder seed",
                "default": None,
            },
            {
                "key": "student_value",
                "description": "the value of the student seed",
                "default": None,
            },
        ],
    },
    {
        "key": "generate",
        "description": "grading configurations to be passed to Otter Generate as an "\
            "otter_config.json; if false, Otter Generate is disabled",
        "default": False,
    },
    {
        "key": "save_environment",
        "description": "whether to save the student's environment in the log",
        "default": False,
    },
    {
        "key": "variables",
        "description": "a mapping of variable names to type strings for serlizing environments",
        "default": {},
    },
    {
        "key": "ignore_modules",
        "description": "a list of modules to ignore variables from during environment serialization",
        "default": [],
    },
    {
        "key": "files",
        "description": "a list of other files to include in the output directories and autograder",
        "default": [],
    },
    {
        "key": "autograder_files",
        "description": "a list of other files only to include in the autograder",
        "default": [],
    },
    {
        "key": "plugins",
        "description": "a list of plugin names and configurations",
        "default": [],
    },
    {
        "key": "test_files",
        "description": "whether to store tests in separate .py files rather than in the notebook " \
            "metadata",
        "default": False,
    },
    {
        "key": "colab",
        "description": "whether this assignment will be run on Google Colab",
        "default": False,
    },
    {
        "key": "show_question_points",
        "description": "whether to add the question point values to the last cell of each question",
        "default": False,
    },
]


class Assignment:
    """
    A class that houses configurations for an assignment. Contains a dictionary of default arguments
    that can be updated in an instance using the ``update()`` method. Functions similarly to an 
    ``AttrDict`` in that keys of the configuration can be accessed as ``assignment.<key>``.

    To access a configuration value, use the dot syntax. For example, to access the ``generate`` key
    of an ``Assignment`` instance ``assignment``:
    
    .. code-block::python
        
        assignment.generate
    
    If ``generate`` is present in ``assignment.config``, then the value in that dictionary will be 
    returned. If it is not, the value in ``Assignment.defaults`` will be returned instead. Configurations
    can also be updated using dot syntax:
    
    .. code-block:: python
        
        assignment.generate = True

    If a key not present in ``Assignment.defaults`` is attempted to be accessed or set, an 
    ``AttributeError`` will be thrown.

    Attributes:
        config (``dict``): the configurations specific to this assignment; keys in this dictionary
            are used before the defaults if present.
    """
    defaults = {
        "master": None,
        "result": None,
        "seed_required": False,
        "_otter_config": None,
        "lang": None,
        "_temp_test_dir": None, # path to a temp dir for tests for otter generate
        "notebook_basename": None,
        **convert_config_description_dict(_DEFAULT_ASSIGNMENT_CONFIGURATIONS_WITH_DESCRIPTIONS),
    }

    def __init__(self):
        self.config = type(self).defaults.copy()

    def __getattr__(self, attr):
        if attr in type(self).defaults:
            return self.config.get(attr, type(self).defaults[attr])
        raise AttributeError(f"Assignment has no attribute {attr}")

    def __setattr__(self, attr, value):
        if attr == "config":
            self.__dict__[attr] = value
        elif attr in type(self).defaults:
            self.config[attr] = value
        else:
            raise AttributeError(f"Assignment has no attribute {attr}")

    def update(self, config):
        """
        Updates the configuration stored in this assignment using keys and values in the dictionary
        ``config``

        Args:
            config (``dict``): new configurations
        """
        for k in config.keys():
            if k not in self.allowed_configs:
                raise ValueError(f"Unexpected assignment config: '{k}'")
        self.config.update(config)

    @property
    def is_r(self):
        """
        Whether the language of the assignment is R
        """
        return self.lang == "r"
    
    @property
    def is_python(self):
        """
        Whether the language of the assignment is Python
        """
        return self.lang == "python"

    @property
    def is_rmd(self):
        """
        Whether the input file is an RMarkdown document
        """
        return self.master.suffix.lower() == ".rmd"
    
    @property
    def allowed_configs(self):
        """
        The list of allowed configuration keys
        """
        return type(self).defaults.keys()


def read_assignment_metadata(cell):
    """
    Return assignment metadata from an assignment cell
    
    Args:
        cell (``nbformat.NotebookNode``): the assignment cell
    
    Returns:
        ``dict``: assignment metadata
    """
    source = get_source(cell)
    begin_assignment_line = get_spec(source, "assignment")
    i, lines = begin_assignment_line + 1, []
    while source[i].strip() != BLOCK_QUOTE:
        lines.append(source[i])
        i = i + 1
    metadata = yaml.full_load('\n'.join(lines))
    return metadata


def is_assignment_cell(cell):
    """
    Returns whether cell contains BEGIN ASSIGNMENT in a block quote
    
    Args:
        cell (``nbformat.NotebookNode``): notebook cell
    
    Returns:
        ``bool``: whether the current cell is an assignment definition cell
    """
    if cell.cell_type != 'markdown':
        return False
    return get_spec(get_source(cell), "assignment") is not None



"""Solution removal for Otter Assign for R notebooks"""

import copy
import re

from ..constants import MD_RESPONSE_CELL_SOURCE
from ..utils import get_source

solution_assignment_regex = re.compile(r"(\s*[\w. $()]*(=|<-))(.*) ?# ?SOLUTION")
def solution_assignment_sub(match):
    """
    Substitutes the first matching group  with ` ...`
    """
    prefix = match.group(1)
    return prefix + ' ...'


solution_line_regex = re.compile(r"(\s*).* ?# ?SOLUTION")
def solution_line_sub(match):
    """
    Substitutes the first matching group  with `# YOUR CODE HERE`
    """
    prefix = match.group(1)
    return prefix + '# YOUR CODE HERE'


SUBSTITUTIONS = [
    (solution_assignment_regex, solution_assignment_sub),
    (solution_line_regex, solution_line_sub),
]



"""Miscellaneous cell generators for Otter Assign"""

import copy
import nbformat

from .constants import MD_RESPONSE_CELL_SOURCE
from .utils import get_source, lock


def gen_init_cell(nb_name, colab):
    """
    Generates a cell to initialize Otter in the notebook.

    Args:
        nb_name (``str``): the name of the notebook being graded
        colab (``bool``): whether the notebook will be run on Colab
    
    Returns:
        ``nbformat.NotebookNode``: the init cell
    """
    if colab:
        args = "colab=True"
    else:
        args  = f"\"{nb_name}\""
    contents = f'# Initialize Otter\nimport otter\ngrader = otter.Notebook({args})'
    cell = nbformat.v4.new_code_cell(contents)
    lock(cell)
    return cell


def gen_markdown_response_cell():
    """
    Generates a Markdown response cell with the following contents:

    .. code-block:: markdown

        _Type your answer here, replacing this text._

    Returns:
        ``nbformat.NotebookNode``: the response cell
    """
    return nbformat.v4.new_markdown_cell(MD_RESPONSE_CELL_SOURCE)


def gen_export_cells(instruction_text, pdf=True, filtering=True, force_save=False, run_tests=False):
    """
    Generates export cells that instruct the student the run a code cell calling 
    ``otter.Notebook.export`` to generate and download their submission. The Markdown cell contains:

    .. code-block:: markdown

        ## Submission
        
        Make sure you have run all cells in your notebook in order before running the cell below, so 
        that all images/graphs appear in the output. The cell below will generate a zipfile for you 
        to submit. **Please save before exporting!**

    Additional instructions can be appended to this cell by passing a string to ``instruction_text``.

    The code cell contains:

    .. code-block:: python

        # Save your notebook first, then run this cell to export your submission.
        grader.export()
    
    The call to ``grader.export()`` contains different arguments based on the values passed to ``pdf``
    and ``filtering``. 
    
    Args:
        instruction_text (``str``): extra instructions for students when exporting
        pdf (``bool``, optional): whether a PDF is needed
        filtering (``bool``, optional): whether PDF filtering is needed
        force_save (``bool``, optional): whether or not to set the ``force_save`` argument of 
            ``otter.Notebook.export`` to ``True``
        run_tests (``bool``, optional): whether to set the ``run_tests`` argument of 
            ``otter.Notebook.export`` to ``True``
    
    Returns:
        ``list`` of ``nbformat.NotebookNode``: generated export cells
    """
    instructions = nbformat.v4.new_markdown_cell()
    instructions.source = "## Submission\n\nMake sure you have run all cells in your notebook in order before " \
    "running the cell below, so that all images/graphs appear in the output. The cell below will generate " \
    "a zip file for you to submit."

    # only include save text if force_save is false
    if not force_save:
        instructions.source += " **Please save before exporting!**"
    
    if instruction_text:
        instructions.source += '\n\n' + instruction_text

    export = nbformat.v4.new_code_cell()
    source_lines = []

    # only include save text if force_save is false
    if not force_save:
        source_lines.append("# Save your notebook first, then run this cell to export your submission.")

    args = []
    if not filtering:
        args += ["filtering=False"]
    elif not pdf:
        args += ["pdf=False"]
    if force_save:
        args += ["force_save=True"]
    if run_tests:
        args += ["run_tests=True"]

    source_lines.append(f"grader.export({', '.join(args)})")
    export.source = "\n".join(source_lines)

    lock(instructions)
    lock(export)

    return [instructions, export, nbformat.v4.new_markdown_cell(" ")]     # last cell is buffer


def gen_check_all_cell():
    """
    Generates a check-all cell and a Markdown cell with instructions to run all tests in the notebook. 
    The Markdown cell has the following contents:

    .. code-block:: markdown

        ---
        
        To double-check your work, the cell below will rerun all of the autograder tests.

    The code cell has the following contents:

    .. code-block:: python

        grader.check_all()
    
    Returns:
        ``list`` of ``nbformat.NotebookNode``: generated check-all cells
    """
    instructions = nbformat.v4.new_markdown_cell()
    instructions.source = "---\n\nTo double-check your work, the cell below will rerun all of the autograder tests."

    check_all = nbformat.v4.new_code_cell("grader.check_all()")

    lock(instructions)
    lock(check_all)

    return [instructions, check_all]


def add_export_tag_to_cell(cell, end=False):
    """
    Adds an HTML comment to close question export for PDF filtering to the top of ``cell``. ``cell``
    should be a Markdown cell. This adds ``<!-- END QUESTION-->`` as the first line of the cell.
    
    Args:
        cell (``nbformat.NotebookNode``): the cell to add the close export to

    Returns:
        ``nbformat.NotebookNode``: the cell with the close export comment at the top
    """
    cell = copy.deepcopy(cell)
    source = get_source(cell)
    tag = "<!-- " + ("END" if end else "BEGIN") + " QUESTION -->"
    source = [tag, ""] + source
    cell['source'] = "\n".join(source)
    return cell



"""
Plugin for using the Gmail API to notify students about the results of public test cases
"""

import os
import json
import tempfile
import gspread
import pandas as pd
import base64
import google.oauth2.credentials

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from googleapiclient.discovery import build
from jinja2 import Template
from textwrap import dedent

from ... import PluginCollection
from ...abstract_plugin import AbstractOtterPlugin


SCOPES = ['https://www.googleapis.com/auth/gmail.send']


class GmailNotifications(AbstractOtterPlugin):
    """
    Otter plugin for sending students an email notification with the results of their public test
    cases. Uses the Gmail API along with user-specified credentials a Google OAuth2 Client to send
    the emails once grading is finished.

    This plugin requires the following configurations:

    - ``client_id``: the Google OAuth2 Client ID
    - ``client_secret``: the Google OAuth2 Client secret
    - ``refresh_tokem``: the Google OAuth2 Client refresh token; more information below
    - ``email``: the email address to send from

    It also supports the following optional configurations:

    - ``catch_api_error``: if this configuration is true, errors thrown by Google API calls are ignored;
      if false, the errors are raised and the grading process stops; defaults to ``True``
    """

    IMPORTABLE_NAME = "otter.plugins.builtin.GmailNotifications"

    email_template_html = Template(dedent("""\
        <p>Hello {{ student_name }},</p>

        <p>Your submission {% if zip_name %}<code>{{ zip_name }}</code>{% endif %} for assignment 
        <strong>{{ assignment_title }}</strong> submitted at {{ submission_timestamp }} received the 
        following scores on public tests:</p>

        <pre>{{ score_report }}</pre>

        <p>If you have any questions or notice any issues, please contact your instructors.</p>

        <hr>

        <p>This message was automatically generated by Otter-Grader.</p>
    """))

    email_template_plain = Template(dedent("""\
        Hello {{ student_name }},

        Your submission {% if zip_name %}{{ zip_name }}{% endif %} for assignment {{ assignment_title }} 
        submitted at {{ submission_timestamp }} received the following scores on public tests:

        {{ score_report }}

        If you have any questions or notice any issues, please contact your instructors.

        ---

        This message was automatically generated by Otter-Grader.
    """))

    def _authenticate(self):
        """
        Create a Google API service using the client ID, secret, and refresh token in the plugin
        configurations.
        """
        try:
            credentials = google.oauth2.credentials.Credentials(
                'token',
                refresh_token=self.plugin_config["refresh_token"],
                token_uri='https://accounts.google.com/o/oauth2/token',
                client_id=self.plugin_config["client_id"],
                client_secret=self.plugin_config["client_secret"],
            )
            self._service = build('gmail', 'v1', credentials=credentials, cache_discovery=False)

        except Exception as e:
            if self.plugin_config.get("catch_api_error", True):
                print(f"Error encountered while authenticating with Gmail:\n{e}")
            else:
                raise e

    @property
    def service(self):
        """
        The Google API service
        """
        if not hasattr(self, "_service") or self._service is None:
            self._authenticate()
        return self._service

    def after_grading(self, results):
        """
        Summarizes the results of grading and formats an email to send to the student using the
        submission metadata. If not submission metadata is supplied, no action is taken.

        Args:
            results (``otter.test_files.GradingResults``): the results of grading
        """
        from ....check.notebook import _ZIP_NAME_FILENAME
        if os.path.exists(_ZIP_NAME_FILENAME):
            with open(_ZIP_NAME_FILENAME) as f:
                zip_name = f.read().strip()
        else:
            zip_name = None
        if self.submission_metadata:
            for user in self.submission_metadata["users"]:
                email_contents = self.email_template_plain.render({
                    "student_name": user["name"],
                    "assignment_title": self.submission_metadata["assignment"]["title"],
                    "submission_timestamp": self.submission_metadata["created_at"],
                    "score_report": results.summary(public_only=True),
                    "zip_name": zip_name,
                })
                plain_email = MIMEText(email_contents, _subtype="plain")

                email_contents = self.email_template_html.render({
                    "student_name": user["name"],
                    "assignment_title": self.submission_metadata["assignment"]["title"],
                    "submission_timestamp": self.submission_metadata["created_at"],
                    "score_report": results.summary(public_only=True),
                    "zip_name": zip_name,
                })
                html_email = MIMEText(email_contents, _subtype="html")

                message = MIMEMultipart('alternative')
                message.attach(plain_email)
                message.attach(html_email)
                message["subject"] = f'Autograder Results for { self.submission_metadata["assignment"]["title"] }'
                message["to"] = user["email"]
                message["from"] = self.plugin_config["email"]
                msg = {'raw': base64.urlsafe_b64encode(message.as_bytes()).decode()}

                try:
                    message = self.service.users().messages().send(userId='me', body=msg).execute()
                except Exception as e:
                    if self.plugin_config.get("catch_api_error", True):
                        print(f'An error occurred while emailing results:\n{e}')
                    else:
                        raise e



"""PDF via LaTeX exporter"""

import nbconvert
import os
import pkg_resources
import warnings

from .base_exporter import BaseExporter, NBCONVERT_6, TEMPLATE_DIR

from ...utils import print_full_width


class PDFViaLatexExporter(BaseExporter):
    """
    Exports notebooks to PDF files using LaTeX as an intermediary

    Converts IPython notebooks to PDFs by first converting them into temporary TeX files that are then
    converted to PDFs using nbconvert and pandoc. Pagebreaks, if enabled, are enforced with a custom
    LaTeX template that clears the document to the next odd numbered page, resulting in responses that
    are all two pages long.

    Attributes:
        default_options (``dict``): the default options for this exporter
    """

    default_options = BaseExporter.default_options.copy()
    default_options.update({
        "save_tex": False,
        "template": "via_latex",
    })

    @classmethod
    def convert_notebook(cls, nb_path, dest, xecjk=False, no_xecjk=False, **kwargs):
        warnings.filterwarnings("ignore", r"invalid escape sequence '\\c'", DeprecationWarning)

        if xecjk and no_xecjk:
            raise ValueError("xeCJK LaTeX template indicated but disallowed")

        options = cls.default_options.copy()
        options.update(kwargs)

        # choose the template to allow Chinese, Japanese, and Korean characters if necessary
        if xecjk:
            options["template"] = "via_latex_xecjk"

        nb = cls.load_notebook(nb_path, filtering=options["filtering"], pagebreaks=options["pagebreaks"])

        if NBCONVERT_6:
            nbconvert.TemplateExporter.extra_template_basedirs = [TEMPLATE_DIR]
            orig_template_name = nbconvert.TemplateExporter.template_name
            nbconvert.TemplateExporter.template_name = options["template"]

        if options["save_tex"]:
            latex_exporter = nbconvert.LatexExporter()
            if not NBCONVERT_6:
                latex_exporter.template_file = os.path.join(TEMPLATE_DIR, options["template"] + ".tpl")

        success = False
        pdf_exporter = nbconvert.PDFExporter()
        if not NBCONVERT_6:
            pdf_exporter.template_file = os.path.join(TEMPLATE_DIR, options["template"] + ".tpl")

        try:
            pdf_output = pdf_exporter.from_notebook_node(nb)
            with open(dest, "wb") as output_file:
                output_file.write(pdf_output[0])

            success = True

            if options["save_tex"]:
                latex_output = latex_exporter.from_notebook_node(nb)
                with open(os.path.splitext(dest)[0] + ".tex", "w+") as output_file:
                    output_file.write(latex_output[0])

            if NBCONVERT_6:
                nbconvert.TemplateExporter.template_name = orig_template_name

        except nbconvert.pdf.LatexFailed as error:
            if not xecjk and not no_xecjk:
                success = cls.convert_notebook(nb_path, dest, xecjk=True, **kwargs)

            elif not success:
                print("There was an error generating your LaTeX; showing full error message:")
                print_full_width("=")
                print(error.output)
                print_full_width("=")
                if xecjk:
                    print("If the error above is related to xeCJK or fandol in LaTeX and you don't "
                        "require this functionality, try running again with no_xecjk set to True "
                        "or the --no-xecjk flag.")

                if NBCONVERT_6:
                    nbconvert.TemplateExporter.template_name = orig_template_name

        return success



"""
OK-formatted test parsers and builders for Otter Assign
"""

import re
import pprint
import yaml
import nbformat

from collections import namedtuple

from .constants import BEGIN_TEST_CONFIG_REGEX, END_TEST_CONFIG_REGEX, TEST_REGEX, OTTR_TEST_NAME_REGEX, \
    OTTR_TEST_FILE_TEMPLATE
from .utils import get_source, lock, str_to_doctest
from ..test_files.abstract_test import TestFile
from ..test_files.metadata_test import NOTEBOOK_METADATA_KEY


Test = namedtuple('Test', ['input', 'output', 'hidden', 'points', 'success_message', 'failure_message'])


def any_public_tests(test_cases):
    """
    Returns whether any of the ``Test`` named tuples in ``test_cases`` are public tests.

    Args:
        test_cases (``list`` of ``Test``): list of test cases
    
    Returns:
        ``bool``: whether any of the tests are public
    """
    return any(not test.hidden for test in test_cases)


def read_test(cell, question, assignment):
    """
    Returns the contents of a test as an ``(input, output, hidden, points, success_message, 
    failure_message)`` named tuple
    
    Args:
        cell (``nbformat.NotebookNode``): a test cell
        question (``dict``): question metadata
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``Test``: test named tuple
    """
    source = get_source(cell)
    if source[0].lstrip().startswith("#"):
        hidden = bool(re.match(r"#\s+hidden\s*", source[0], flags=re.IGNORECASE))
    else:
        hidden = False

    i = 0 if hidden else -1

    output = ''
    for o in cell['outputs']:
        output += ''.join(o.get('text', ''))
        results = o.get('data', {}).get('text/plain')
        if results and isinstance(results, list):
            output += results[0]
        elif results:
            output += results

    if re.match(BEGIN_TEST_CONFIG_REGEX, source[0], flags=re.IGNORECASE):
        for i, line in enumerate(source):
            if re.match(END_TEST_CONFIG_REGEX, line, flags=re.IGNORECASE):
                break
        config = yaml.full_load("\n".join(source[1:i]))
        assert isinstance(config, dict), f"Invalid test config in cell {cell}"
    else:
        config = {}
    
    hidden = config.get("hidden", hidden)
    points = config.get("points", None)
    success_message = config.get("success_message", None)
    failure_message = config.get("failure_message", None)

    return Test('\n'.join(get_source(cell)[i+1:]), output, hidden, points, success_message, failure_message)


def gen_test_cell(question, tests, tests_dict, assignment):
    """
    Parses a list of test named tuples and creates a single test file. Adds this test file as a value
    to ``tests_dict`` with a key corresponding to the test's name, taken from ``question``. Returns
    a code cell that runs the check on this test.
    
    Args:
        question (``dict``): question metadata
        tests (``list`` of ``Test``): tests to be written
        tests_dict (``dict``): the tests for this assignment
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``nbformat.NotebookNode``: code cell calling ``otter.Notebook.check`` on this test
    """
    cell = nbformat.v4.new_code_cell()

    cell.source = ['grader.check("{}")'.format(question['name'])]

    points = question.get('points', None)
    if isinstance(points, dict):
        points = points.get('each', 1) * len(tests)
    elif isinstance(points, list):
        if len(points) != len(tests):
            raise ValueError(
                f"Error in question {question['name']}: length of 'points' is {len(points)} but there "
                f"are {len(tests)} tests")

    # check for errors in resolving points
    TestFile.resolve_test_file_points(points, tests)

    suites = [gen_suite(tests)]
    test = {
        'name': question['name'],
        'points': points,
        'suites': suites,
    }

    tests_dict[question['name']] = test
    lock(cell)
    return cell


def gen_suite(tests):
    """
    Generates an OK test suite for a list of tests as named tuples
    
    Args:
        tests (``list`` of ``otter.assign.Test``): test cases

    Returns:
        ``dict``: OK test suite
    """
    cases = [gen_case(test) for test in tests]
    return  {
      'cases': cases,
      'scored': True,
      'setup': '',
      'teardown': '',
      'type': 'doctest'
    }


def gen_case(test):
    """
    Generates an OK test case for a test named tuple
    
    Args:
        test (``otter.assign.Test``): OK test for this test case

    Returns:
        ``dict``: the OK test case
    """
    code_lines = str_to_doctest(test.input.split('\n'), [])
    code_lines.append(test.output)
    ret = {
        'code': '\n'.join(code_lines),
        'hidden': test.hidden,
        'locked': False,
    }
    if test.points is not None:
        ret['points'] = test.points
    if test.success_message:
        ret['success_message'] = test.success_message
    if test.failure_message:
        ret['failure_message'] = test.failure_message
    return ret


def write_test(nb, path, test, use_file=False):
    """
    Writes an OK test file
    
    Args:
        nb (``dict``): the notebook being written
        path (``str``): path of file to be written or the name of the test
        test (``dict``): OK test to be written
        use_file (``bool``): whether to write to a file instead of putting the test in the notebook
            metadata
    """
    if use_file:
        with open(path, 'w+') as f:
            if isinstance(test, dict):
                f.write('test = ')
                pprint.pprint(test, f, indent=4, width=200, depth=None)
            else:
                f.write(test)
    else:
        if NOTEBOOK_METADATA_KEY not in nb["metadata"]:
            nb["metadata"][NOTEBOOK_METADATA_KEY] = {}
        if "tests" not in nb["metadata"][NOTEBOOK_METADATA_KEY]:
            nb["metadata"][NOTEBOOK_METADATA_KEY]["tests"] = {}
        nb["metadata"][NOTEBOOK_METADATA_KEY]["tests"][path] = test


def remove_hidden_tests_from_dir(nb, test_dir, assignment, use_files=False):
    """
    Rewrites test files in a directory to remove hidden tests
    
    Args:
        nb (``dict``): the notebook being written
        test_dir (``pathlib.Path``): path to test files directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        use_file (``bool``): whether separate test files were written instead of notebook metadata
    """
    if use_files:
        for f in test_dir.iterdir():
            if f.name == '__init__.py' or f.suffix != '.py':
                continue
            locals = {}
            with open(f) as f2:
                exec(f2.read(), globals(), locals)
            test = locals['test']
            for suite in test['suites']:
                for i, case in list(enumerate(suite['cases']))[::-1]:
                    if case['hidden']:
                        suite['cases'].pop(i)
                        if isinstance(test['points'], list):
                            test['points'].pop(i)
            write_test({}, f, test, use_file=True)
    else:
        tests = nb["metadata"].get(NOTEBOOK_METADATA_KEY, {}).get("tests", {})
        for tn, test in tests.items():
            for i, tc in list(enumerate(test["suites"][0]["cases"]))[::-1]:
                if tc["hidden"]:
                    test["suites"][0]["cases"].pop(i)


def determine_question_point_value(question_metadata, test_cases):
    """
    Determine the point value of a question using the question metadata and list of test cases.

    Args:
        question_metadata (``dict[str, object]``): the question metadata
        test_cases (``list[Test]``): the test cases for the question; if a manual question, this list
            should be empty

    Returns:
        number: the point value of the question
    """
    if len(test_cases) == 0:
        if question_metadata["points"] is None and question_metadata["manual"]:
            raise ValueError(f"Point value unspecified for question with no test cases: {question_metadata['name']}")

        return question_metadata["points"] if question_metadata["points"] is not None else 1

    resolved_test_cases = TestFile.resolve_test_file_points(question_metadata["points"], test_cases)
    points = sum(tc.points for tc in resolved_test_cases)
    return int(points) if points % 1 == 0 else points



"""
Gradescope token generator; original script provided by Gradescope
"""

import requests
import getpass

BASE_URL = 'https://www.gradescope.com'

class APIClient:
    """
    Client for Gradescope's API. Logs in users and retrieves token, uploads PDF and code assignment
    submissions. 
    
    Client code originally provided by Gradescope.

    Args:
        token (``str``, optional): a pre-retrieved token for the client

    Attributes:
        token (``str``): the user's Gradescope token
    """
    def __init__(self, token=None):
        self.session = requests.Session()
        if token is not None:
            self.token = token

    @classmethod
    def get_token(cls):
        """
        Prompts a user for their Gradescope username and password and retrieves a token for that user

        Returns:
            ``str``: the user's token
        """
        client = cls()
        email = input("Please provide the email address on your Gradescope account: ")
        password = getpass.getpass('Password: ')
        client.log_in(email, password)
        return client.token

    def post(self, *args, **kwargs):
        return self.session.post(*args, **kwargs)

    def log_in(self, email, password):
        url = "{base}/api/v1/user_session".format(base=BASE_URL)

        form_data = {
            "email": email,
            "password": password
        }
        r = self.post(url, data=form_data)
        assert r.status_code ==  200, r.text

        self.token = r.json()['token']

    def upload_pdf_submission(self, course_id, assignment_id, student_email, filename):
        url = "{base}/api/v1/courses/{0}/assignments/{1}/submissions".format(
            course_id, assignment_id, base=BASE_URL
        )

        form_data = {
            "owner_email": student_email
        }
        files = {'pdf_attachment': open(filename, 'rb')}

        request_headers = {'access-token': self.token}
        r = self.post(url, data=form_data, headers=request_headers, files=files)
        return r

    def replace_pdf_submission(self, course_id, assignment_id, student_email, filename):
        url = "{base}/api/v1/courses/{0}/assignments/{1}/submissions/replace_pdf".format(
            course_id, assignment_id, base=BASE_URL
        )

        form_data = {
            "owner_email": student_email
        }
        files = {'pdf_attachment': open(filename, 'rb')}

        request_headers = {'access-token': self.token}
        r = self.post(url, data=form_data, headers=request_headers, files=files)
        return r

    def upload_programming_submission(self, course_id, assignment_id, student_email, filenames):
        url = "{base}/api/v1/courses/{0}/assignments/{1}/submissions".format(
            course_id, assignment_id, base=BASE_URL
        )

        form_data = {
            "owner_email": student_email
        }
        files = [('files[]', (filename, open(filename, 'rb'))) for filename in filenames]

        request_headers = {'access-token': self.token}
        r = self.post(url, data=form_data, headers=request_headers, files=files)
        return r



"""
Utils for Otter Check
"""

import hashlib
import os
import pickle
import tempfile
import time

from IPython import get_ipython
from IPython.display import display, Javascript
from subprocess import run, PIPE

from .logs import EventType


def save_notebook(filename, timeout=10):
    """
    Force-saves a Jupyter notebook by displaying JavaScript.

    Args:
        filename (``str``): path to notebook file being saved
        timeout (``int`` or ``float``): number of seconds to wait for save before timing-out

    Returns
        ``bool``: whether the notebook was saved successfully
    """
    timeout = timeout * 10**9
    if get_ipython() is not None:
        with open(filename, "rb") as f:
            md5 = hashlib.md5(f.read()).hexdigest()
        start = time.time_ns()
        display(Javascript("Jupyter.notebook.save_checkpoint();"))

        curr = md5
        while curr == md5 and time.time_ns() - start <= timeout:
            time.sleep(1)
            with open(filename, "rb") as f:
                curr = hashlib.md5(f.read()).hexdigest()


        return curr != md5

    return True


def grade_zip_file(zip_path, nb_arcname, tests_dir):
    """
    Grade a submission zip file in a separate process and return the ``GradingResults`` object.
    """
    _, results_path = tempfile.mkstemp(suffix=".pkl")

    try:
        command = [
            "python3", "-m", "otter.check.validate_export",
            "--zip-path", zip_path,
            "--nb-arcname", nb_arcname,
            "--tests-dir", tests_dir,
            "--results-path", results_path,
        ]

        # run the command
        results = run(command, stdout=PIPE, stderr=PIPE)

        if results.stderr:
            raise RuntimeError(results.stderr)

        with open(results_path, "rb") as f:
            results = pickle.load(f)

        return results

    finally:
        os.remove(results_path)


def running_on_colab():
    """
    Determine whether the current environment is running on Google Colab by checking the IPython
    interpreter.

    Returns:
        ``bool``: whether the current environment is on Google Colab
    """
    return "google.colab" in str(get_ipython())


def colab_incompatible(f):
    """
    A decator that raises an error if the wrapped function is called in an environment running on
    Google Colab.
    """
    def colab_only_method(self, *args, **kwargs):
        if self._colab:
            raise RuntimeError("This method is not compatible with Google Colab")
        return f(self, *args, **kwargs)
    return colab_only_method


def logs_event(event_type):
        """
        A decorator that ensures each call is logged in the Otter log with type ``event_type``.

        Events logging a ``EventType.CHECK`` should return a 3-tuple of the question name, the
        ``TestFile`` and an environment to shelve. All other methods should just return their 
        default return value, which will be logged.
        """

        def event_logger(f):
            """
            Wraps a function and ensures that the call's results are logged using 
            ``Notebook._log_event``.
            """

            def run_function(self, *args, **kwargs):
                """
                Runs a method, catching any errors and logging the call. Returns the return value
                of the function, unless ``EventType.CHECK`` is used, in which case the return value
                is assumed to be a 3-tuple and the second value in the tuple is returned.
                """
                try:
                    if event_type == EventType.CHECK:
                        question, results, shelve_env = f(self, *args, **kwargs)
                    else:
                        results = f(self, *args, **kwargs)
                        shelve_env = {}
                        question = None
                except Exception as e:
                    self._log_event(event_type, success=False, error=e)
                    raise e
                else:
                    self._log_event(event_type, results=results, question=question, shelve_env=shelve_env)
                return results

            return run_function

        return event_logger



"""Various constants for Otter Assign"""

import re

from jinja2 import Template
from textwrap import indent


NB_VERSION = 4
BLOCK_QUOTE = "```"
COMMENT_PREFIX = "#"
TEST_HEADERS = ["TEST", "HIDDEN TEST"]
ALLOWED_NAME = re.compile(r'[A-Za-z][A-Za-z0-9_]*')
NB_VERSION = 4

MD_RESPONSE_CELL_SOURCE = "_Type your answer here, replacing this text._"

BEGIN_TEST_CONFIG_REGEX = r'(?:.\s*=\s*)?""?"?\s*#\s*BEGIN\s*TEST\s*CONFIG'
END_TEST_CONFIG_REGEX = r'""?"?\s*#\s*END\s*TEST\s*CONFIG'
TEST_REGEX = rf"(##\s*(hidden\s*)?test\s*##\s*|#\s*(hidden\s*)?test\s*|{BEGIN_TEST_CONFIG_REGEX})"
OTTR_TEST_NAME_REGEX = r'''(?:testthat::)?test_that\(['"]([A-Za-z0-9_]+)['"],'''
MD_SOLUTION_REGEX = r"(<strong>|\*{2})solution:?(<\/strong>|\*{2})"
SEED_REGEX = r"##\s*seed\s*##"
IGNORE_REGEX = r"(##\s*ignore\s*##\s*|#\s*ignore\s*)"

# DON'T change this template or the regex that removes hidden tests will break in the R adapter!
OTTR_TEST_FILE_TEMPLATE = Template("""\
test = list(
  name = "{{ name }}",
  cases = list({% for tc in test_cases %}
    ottr::TestCase$new(
      hidden = {% if tc.hidden %}TRUE{% else %}FALSE{% endif %},
      name = {% if tc.name %}"{{ tc.name }}"{% else %}NA{% endif %},
      points = {% if tc.points is none %}NA{% else %}{{ tc.points }}{% endif %},{% if tc.success_message %}
      success_message = "{{ tc.success_message }}",{% endif %}{% if tc.failure_message %}
      failure_message = "{{ tc.failure_message }}",{% endif %}
      code = {
        {{ indent(tc.body, "        ").lstrip() }}{# lstrip so that the first line indent is correct #}
      }
    ){% if not loop.last %},{% endif %}{% endfor %}
  )
)""")
OTTR_TEST_FILE_TEMPLATE.globals['indent'] = indent



##################################
##### Tests for otter assign #####
##################################

import unittest
import sys
import os
import shutil
import subprocess

from subprocess import PIPE
from glob import glob
from unittest.mock import patch

from otter.assign import main as assign
from otter.generate.token import APIClient

from . import TestCase


TEST_FILES_PATH = "test/test-assign/"


class TestAssign(TestCase):

    def check_gradescope_zipfile(self, path, correct_dir_path):
        # unzip the zipfile
        unzip_command = ["unzip", "-o", path, "-d", TEST_FILES_PATH + "autograder"]
        unzip = subprocess.run(unzip_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(unzip.stderr), 0, unzip.stderr.decode("utf-8"))

        self.assertDirsEqual(TEST_FILES_PATH + "autograder", correct_dir_path)

        # cleanup
        if os.path.exists(TEST_FILES_PATH + "autograder"):
            shutil.rmtree(TEST_FILES_PATH + "autograder")

    def test_convert_example(self):
        """
        Checks that otter assign filters and outputs correctly
        """
        # run otter assign
        assign(TEST_FILES_PATH + "example.ipynb", TEST_FILES_PATH + "output", no_run_tests=True, v1=True)
       
        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "example-correct")
        
    def test_otter_example(self):
        """
        Checks that otter assign filters and outputs correctly, as well as creates a correct .otter file
        """
        assign(TEST_FILES_PATH + "generate-otter.ipynb", TEST_FILES_PATH + "output")
        
        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "otter-correct")     

    def test_pdf_example(self):
        """
        Checks that otter assign filters and outputs correctly, as well as creates a correct .zip file along with PDFs
        """
        assign(TEST_FILES_PATH + "generate-pdf.ipynb", TEST_FILES_PATH + "output", no_run_tests=True)
      
        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "pdf-correct", 
            ignore_ext=[".pdf"], variable_path_exts=[".zip"])
    
    @patch.object(APIClient, "get_token")
    def test_gradescope_example(self, mocked_client):
        """
        Checks that otter assign filters and outputs correctly, as well as creates a correct .zip file along with PDFs.
        Additionally, includes testing Gradescope integration.
        """
        # set a return value that does not match the token in the notebook, so we'll know if APIClient
        # is called
        mocked_client.return_value = 'token'  

        assign(TEST_FILES_PATH + "generate-gradescope.ipynb", TEST_FILES_PATH + "output", no_run_tests=True)
      
        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "gs-correct", 
            ignore_ext=[".pdf"], variable_path_exts=[".zip"])

        # check gradescope zip file
        self.check_gradescope_zipfile(
            glob(TEST_FILES_PATH + "output/autograder/*.zip")[0], 
            TEST_FILES_PATH + "gs-autograder-correct",
        )

    def test_r_example(self):
        """
        Checks that otter assign works for R notebooks correctly
        """
        assign(TEST_FILES_PATH + "r-example.ipynb", TEST_FILES_PATH + "output", v1=True)

        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "r-correct", 
            ignore_ext=[".pdf"], variable_path_exts=[".zip"])

    def test_rmd_example(self):
        """
        Checks that otter assign works for Rmd files
        """
        assign(TEST_FILES_PATH + "rmd-example.Rmd", TEST_FILES_PATH + "output", v1=True)

        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "rmd-correct", 
            ignore_ext=[".pdf"], ignore_dirs=["rmd-example-sol_files"], variable_path_exts=[".zip"])
        
        # check gradescope zip file
        self.check_gradescope_zipfile(
            glob(TEST_FILES_PATH + "output/autograder/*.zip")[0], 
            TEST_FILES_PATH + "rmd-autograder-correct",
        )

    def tearDown(self):
        """
        Removes assign output
        """
        # cleanup
        if os.path.exists(TEST_FILES_PATH + "output"):
            shutil.rmtree(TEST_FILES_PATH + "output")
